import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy.spatial import cKDTree

def coordinates_within_margin_selection(coordinates,  image = None, bounds = None, margin=10):
    if coordinates.size == 0:
        return np.array([])

    if image is not None:
        bounds = np.array([[0,0], [image.shape[1],image.shape[0]]])

    criteria = np.array([(coordinates[:, 0] > (bounds[0,0] + margin)),
                         (coordinates[:, 0] < (bounds[1,0] - margin)),
                         (coordinates[:, 1] > (bounds[0,1] + margin)),
                         (coordinates[:, 1] < (bounds[1,1] - margin))
                         ])

    return criteria.all(axis=0)

def coordinates_within_margin_selection(coordinates,  image = None, bounds = None, margin=10):
    if coordinates.size == 0:
        return np.array([])

    if image is not None:
        bounds = np.array([[0,0], [image.shape[1],image.shape[0]]])

    if isinstance(margin, int):
        margin = np.array([margin, margin])

    criteria = np.array([(coordinates[:, 0] > (bounds[0, 0] + margin[0])),
                         (coordinates[:, 0] < (bounds[1, 0] - margin[0])),
                         (coordinates[:, 1] > (bounds[0, 1] + margin[1])),
                         (coordinates[:, 1] < (bounds[1, 1] - margin[1]))
                         ])

    return criteria.all(axis=0)


def coordinates_within_margin(coordinates, image=None, bounds=None, margin=10):
    criteria = coordinates_within_margin_selection(coordinates,  image=image, bounds=bounds, margin=margin)
    if criteria.size == 0:
        return np.array([])
    else:
        return coordinates[criteria]


def circle(r):
    d = 2*r + 1
    rx, ry = d/2, d/2
    x, y = np.indices((d, d))
    return (np.abs(np.hypot(r - x, r - y)-r) < 0.5).astype(int)

def coordinates_without_intensity_at_radius(coordinates, image, radius, cutoff, fraction_of_peak_max = 0.25):
    if cutoff == 'image_median': cutoff = np.median(image)
    circle_matrix = circle(radius)
    new_coordinates = []

    coordinates = coordinates_within_margin(coordinates, image=image, margin=radius+1)

    for i, coordinate in enumerate(coordinates):
        # Could use the coordinates_within_margin for this [IS 01-11-2019]
        #if np.all(coordinate > radius+1) and \
        #        np.all(coordinate < (np.array(image.shape)-radius-1)):

        cropped_peak = crop(image, coordinate, radius*2+1)
        #if np.all((cropped_peak * circle_matrix) < (cutoff + fraction_of_peak_max * np.max(cropped_peak))): #OLd version!
        if np.all((cropped_peak * circle_matrix) < (cutoff + fraction_of_peak_max * (np.max(cropped_peak)-cutoff))):
            new_coordinates.append(coordinate)

    return np.array(new_coordinates)

def crop(image, center, width):
    center = np.round(center).astype(int)
    return image[(center[1]-width//2):(center[1]+width//2+1),(center[0]-width//2):(center[0]+width//2+1)]

# def twoD_gaussian(M, offset, amplitude, x0, y0, sigma_x, sigma_y):
#     x, y = M
#     return offset + amplitude * np.exp(- ((x-x0)/(2*sigma_x))**2 - ((y-y0)/(2*sigma_y))**2)

def twoD_gaussian(M, offset, amplitude, x0, y0, sigma):
    x, y = M
    return offset + amplitude * np.exp(- ((x-x0)**2+(y-y0)**2)/(2*sigma**2))

def fit_twoD_gaussian(Z):
    height, width = Z.shape
    x, y = np.arange(width)-width//2, np.arange(height)-height//2
    X, Y = np.meshgrid(x, y)
    xdata = np.vstack((X.ravel(), Y.ravel()))

    # p0 = [20,20,0,0,1,1]
    # p0 = [np.min(Z), np.max(Z) - np.min(Z), 0, 0, 1, 1]
    p0 = [np.min(Z), np.max(Z) - np.min(Z), 0, 0, 1]
    popt, pcov = curve_fit(twoD_gaussian, xdata, Z.ravel(), p0) #, maxfev=3000) #input: function, xdata, ydata,p0

    # The offset can potentially be used for background subtraction
    return popt

def coordinates_after_gaussian_fit(coordinates, image, gaussian_width = 9, return_fit_parameters=False):
    # TODO: fix standard deviation to the psf size, this may improve correct peak localization when two peaks are close.
    new_coordinates = []
    fit_parameters = []
    if len(coordinates) == 0:  # This statement may not be necessary. However, check the code thoroughly before you remove this.
        new_coordinates = coordinates
    else:
        coordinates = coordinates_within_margin(coordinates, image=image, margin=gaussian_width//2+1)

        for i, coordinate in enumerate(coordinates):
            # Could use the coordinates_within_margin for this [IS 01-11-2019]
            #if np.all(coordinate > gaussian_width//2+1) and \
            #        np.all(coordinate < np.array(image.shape)-gaussian_width//2-1):
            cropped_peak = crop(image, coordinate, gaussian_width)
            try:
                coefficients = fit_twoD_gaussian(cropped_peak)
                #new_coordinates.append(coordinate + coefficients[2:4])
                new_coordinate = coordinate + coefficients[2:4]
                if np.sum(np.abs(coefficients[2:4]))<gaussian_width*2:
                    new_coordinates.append(new_coordinate)
                    fit_parameters.append(coefficients)
                # else: #MD do nothing, you don't want to include fits with a center far outside the cropped image
            except RuntimeError:
                pass
    if return_fit_parameters:
        return np.array(new_coordinates), np.array(fit_parameters)
    else:
        return np.array(new_coordinates)


def merge_nearby_coordinates(coordinates, distance_threshold=2, plot=False):
    """Merge nearby coordinates to a single coordinate

    Coordinates are stored in a KD-tree.
    Each pair of points with a distance smaller than the distance threshold is obtained
    Pairs are chained to obtain groups of points
    For each group find the center coordinate and use that as a new coordinate
    (do this only if each member of the group is within the distance threshold from the center coordinate).
    Add individual points to the new coordinate list, i.e. points that do not have other points within the distance threshold.

    Parameters
    ----------
    coordinates : numpy.ndarray of ints or floats OR set of tuples
        Array with each row a set of coordinates
    distance_threshold : int or float
        Points closer than this distance are considered belonging to the same molecule.
    plot : bool
        If True shows a scatter plot of the coordinates and the new coordinates on top. (Only for 2D coordinates)

    Returns
    -------
    new_coordinates : numpy.ndarray of floats
        Coordinate array after merging nearby coordinates

    """

    # Convert to numpy array in case the coordinates are given as a set of tuples
    coordinates = array_from_set_of_tuples(coordinates)

    # Put coordinates in KD-tree for fast nearest-neighbour finding
    coordinates_KDTree = cKDTree(coordinates)

    # Determine pairs of points closer than the distance_threshold
    close_pairs = coordinates_KDTree.query_pairs(r=distance_threshold)
    close_pairs = [set(pair) for pair in close_pairs] # Convert to list of sets

    # Chain the pairs to obtain groups (or clusters) of points
    groups_of_points = combine_overlapping_sets(close_pairs)

    # Calculate the new coordinates by taking the center of all the neighbouring points.
    # A threshold for the total group is applied, i.e. all points must lie within the distance_threshold
    # from the center coordinate.
    new_coordinates = []
    for group in groups_of_points:
        group_coordinates = coordinates[list(group)]
        center_coordinate = np.mean(group_coordinates, axis=0)
        distances_to_center = np.sqrt(np.sum((group_coordinates-center_coordinate)**2, axis=1))
        if not (np.max(distances_to_center) > distance_threshold): # This could be another threshold
            new_coordinates.append(center_coordinate)

    # Obtain individual points, i.e. those that do not have another point within the distance_threshold.
    # This is done by taking the difference from all points and the ones that are present in any of the groups.
    all_points_in_groups = set(point for group in groups_of_points for point in group)
    all_points = set(range(len(coordinates)))
    individual_points = all_points.difference(all_points_in_groups)

    # Add individual points to new_coordinates list
    for point in individual_points:
        new_coordinates.append(coordinates[point])

    # Convert to numpy array
    new_coordinates = np.array(new_coordinates)

    if plot:
        axis = plt.figure().gca()
        axis.scatter(coordinates[:,0],coordinates[:,1])
        axis.scatter(new_coordinates[:,0],new_coordinates[:,1])

    return new_coordinates


def combine_overlapping_sets(old_list_of_sets):
    """ Combine sets that have overlap

    Go through each set, if it has overlap with one of the sets in the new list of sets, then combine it with this set
    If there is no overlap, append the set to the new list of sets.
    Perform this function recursively until the new_list_of_sets does not change anymore.

    Parameters
    ----------
    old_list_of_sets : list of sets
        List of sets of which overlapping ones should be combined

    Returns
    -------
    new_list_of_sets : list of sets
        Combined list of sets

    """

    # test_set1 = [set((1,2)),set((3,4)),set((5,2)),set((5,6)),set((4,10))]
    # test_set2 = [set((1,2)),set((3,4)),set((2,3)),set((5,6)),set((4,10))]

    new_list_of_sets = []
    for old_set in old_list_of_sets:
        append = True
        for new_set in new_list_of_sets:
            if not (old_set.isdisjoint(new_set)):
                new_set.update(old_set)
                append = False
        if append:
            new_list_of_sets.append(old_set.copy())

    if not (new_list_of_sets == old_list_of_sets):
        new_list_of_sets = combine_overlapping_sets(new_list_of_sets)

    return new_list_of_sets


def set_of_tuples_from_array(array):
    return set([tuple(a) for a in array])


def array_from_set_of_tuples(set_of_tuples):
    return np.array([t for t in set_of_tuples])



if __name__ == '__main__':
    from papylio import Experiment
    from papylio.peak_finding import find_peaks

    exp = Experiment(r'D:\ivoseverins\SURFdrive\Promotie\Code\Python\traceAnalysis\twoColourExampleData\20141017 - Holliday junction - Copy\HJC-50pM')
    file = exp.files[2]
    movie = file.movie
    image = movie.make_average_image(write=False)

    coordinates = find_peaks(image=image, method='adaptive-threshold', minimum_area=5, maximum_area=15)

    plt.imshow(image)
    plt.scatter(coordinates[:, 0], coordinates[:, 1], color='b')

    coordinates = coordinates_within_margin(coordinates, bounds = np.array([[0,255], [0,511]]), margin=10)
    plt.scatter(coordinates[:, 0], coordinates[:, 1], color='g')

    coordinates = coordinates_after_gaussian_fit(coordinates, image)
    plt.scatter(coordinates[:, 0], coordinates[:, 1], color='y')

    coordinates = coordinates_without_intensity_at_radius(coordinates,
                                                          image,
                                                          radius=4,
                                                          cutoff=np.median(image),
                                                          fraction_of_peak_max=0.35) # was 0.25 in IDL code
    plt.scatter(coordinates[:, 0], coordinates[:, 1], color='r')



import itertools
import numpy as np
import matplotlib.pyplot as plt
import xarray as xr
import seaborn as sns
sns.stripplot
def basepairing(sequences):
    s = sequences.astype('U').view('U1').reshape(-1, 8)
    s1 = xr.DataArray(s, coords=[('sequence', sequences), ('nucleotide_1', range(8))])
    s2 = xr.DataArray(s, coords=[('sequence', sequences), ('nucleotide_2', range(8))])
    s11, s22 = xr.broadcast(s1, s2)
    basepairs = xr.concat([s11,s22], dim='b').str.join(dim='b')
    basepairing = (basepairs == 'AT').astype(int) + (basepairs == 'TA').astype(int) + (basepairs == 'CG').astype(int) + (basepairs == 'GC').astype(int)
    return basepairing


def plot_basepairing(basepairing, title='', ax=None, save_path=None, max_linewidth=10):
    data = basepairing.mean('sequence')

    n = 8
    phi = 9/8*np.pi
    angle = np.linspace(0,2*np.pi,n+1)[:n]
    coordinates = np.vstack([np.sin(angle+phi), np.cos(angle+phi)]).T

    if ax is None:
        fig, ax = plt.subplots(tight_layout=True)
    else:
        fig = ax.figure
    ax.set_aspect(1)
    ax.axes.set_axis_off()
    ax.set_xlim(-1.1, 1.1)
    ax.set_ylim(-1.1, 1.1)
    for i,j in itertools.combinations(range(n),2):
        # ax.annotate("",
        #             xy=coordinates[i], xycoords='data',
        #             xytext=coordinates[j], textcoords='data',
        #             arrowprops=dict(arrowstyle="-"),
        #             )
        ax.plot(*coordinates[[i,j]].T, c='k', linewidth=data[i,j]*max_linewidth)
    ax.scatter(*coordinates.T, s=400, c='k')
    for i, c in enumerate(coordinates):
        ax.annotate(i, c, ha='center', va='center', c='white')
        # ax.text(*c, i + 1, size=15, ha='center', va='center', c='white',
        #         bbox=dict(boxstyle="circle,pad=0.3", fc="cyan", ec="b", lw=2))
    ax.set_title(title)

    if save_path is not None:
        fig.savefig(save_path.joinpath(f'Basepairing - '+ title +'.png'))


def plot_basepairing_individual(sequences, name='', rows=5, columns=10, titles=None, save_path=None, max_linewidth=5):
    fig, axes = plt.subplots(rows, columns, figsize=(columns*2.5, rows*2.5))#, tight_layout=True)
    fig.subplots_adjust(hspace=0.3, wspace=0.3, left=0.02, right=0.98, bottom=0.02, top=0.95)
    axes = axes.flatten()

    for ax in axes:
        ax.set_axis_off()

    for i, (sequence, ax) in enumerate(zip(sequences, axes)):
        if titles is not None:
            title = f'{titles[i]} - {sequence}'
        else:
            title = f'{sequence}'
        plot_basepairing(basepairing(np.array([sequence])), title=title, ax=ax, max_linewidth=max_linewidth)
    if save_path is not None:
        fig.savefig(save_path.joinpath(f'Basepairing - {name} - individual.png'))


def basepairing_possible(sequences):
    bp = basepairing(sequences)
    return basepairing



def plot_basepairing(basepairing, title='', ax=None, save_path=None, max_linewidth=10):
    data = basepairing.mean('sequence')

    n = 8
    phi = 9/8*np.pi
    angle = np.linspace(0,2*np.pi,n+1)[:n]
    coordinates = np.vstack([np.sin(angle+phi), np.cos(angle+phi)]).T

    if ax is None:
        fig, ax = plt.subplots(tight_layout=True)
    else:
        fig = ax.figure
    ax.set_aspect(1)
    ax.axes.set_axis_off()
    ax.set_xlim(-1.1, 1.1)
    ax.set_ylim(-1.1, 1.1)
    for i,j in itertools.combinations(range(n),2):
        # ax.annotate("",
        #             xy=coordinates[i], xycoords='data',
        #             xytext=coordinates[j], textcoords='data',
        #             arrowprops=dict(arrowstyle="-"),
        #             )
        ax.plot(*coordinates[[i,j]].T, c='k', linewidth=data[i,j]*max_linewidth)
    ax.scatter(*coordinates.T, s=400, c='k')
    for i, c in enumerate(coordinates):
        ax.annotate(i, c, ha='center', va='center', c='white')
        # ax.text(*c, i + 1, size=15, ha='center', va='center', c='white',
        #         bbox=dict(boxstyle="circle,pad=0.3", fc="cyan", ec="b", lw=2))
    ax.set_title(title)

    if save_path is not None:
        fig.savefig(save_path.joinpath(f'Basepairing - '+ title +'.png'))


# -*- coding: utf-8 -*-
"""
Created on Mon Jul 13 15:50:57 2020

@author: mwdoc
https://github.com/soft-matter/pims_nd2
read_header and read_frame adapted, def __init__ unchanged
"""

from pathlib import Path
import os, sys

import time
import numpy as np
import matplotlib.pyplot as plt
import xarray as xr
from nd2reader import ND2Reader

from papylio.movie.movie import Movie, Illumination


class ND2Movie(Movie):
    extensions = ['.nd2']

    def __init__(self, arg, *args, **kwargs):
        super().__init__(arg, *args, **kwargs)
        # super().__init__(arg)

        self.writepath = self.filepath.parent
        self.name = self.filepath.with_suffix('').name

        if 'fov' in self.name:
            token_position = self.name.find('_fov')
            self.fov_index = int(self.name[token_position+4:])
            self.name = self.name[:token_position]
            self.filepath = self.filepath.with_name(self.name).with_suffix('.nd2')
        else:
            self.fov_index = None

        # setting for multi fov measurement
        # # TODO: Move fov info / fov selection here if possible
        # self.fov_info = None
        # if 'fov_info' in kwargs:
        #     self.fov_info = kwargs['fov_info']  # fov=Field of View

        self.threshold = {'view': (0, 200),
                          'point-selection': (45, 25)
                          }

        # We should probably put this in the configuration file
        # SHK: self.rot90 should be set before reading the header.
        # self.rot90 = 1

        # self.read_header()

        # self.time = self.time[self.fov_info['first_frame_of_each_fov'][self.fov_info['fov_chosen']]:(self.fov_info['last_frame_of_each_fov'][self.fov_info['fov_chosen']]+1)]
        # self.illumination = self.illumination[self.fov_info['first_frame_of_each_fov'][self.fov_info['fov_chosen']]:(self.fov_info['last_frame_of_each_fov'][self.fov_info['fov_chosen']]+1)]
        # self.create_frame_info()  # Possibly move to Movie later on

        # self._initialized = True

        self.file = None # Note this is for the tif file, not the File class.

    def open(self):
        self.file = ND2Reader(str(self.filepath))
        if 'c' in self.file.axes:
            self.file.iter_axes = 'tc'
        else:
            self.file.iter_axes = 't'

    def close(self):
        self.file.close()

    def _read_header(self):
        with self:
            y_positions = self.file._parser._raw_metadata.y_data  # nikon sample stage position
            x_positions = self.file._parser._raw_metadata.x_data  # nikon sample stage position

            # TODO: @Sung Hyun, I assume it doe snot matter for y and x-positions if self.file.iter_axis is called first, right?
            # Then we can remove the commented lines below
            # # set the image data order in the nd2 file
            # if 'c' in self.file.axes:
            #     self.file.iter_axes = 'tc'  # for alex measurements
            # else:
            #     self.file.iter_axes = 't'

            n_illumination = len(self.file.metadata["channels"])
            n_frames = len(x_positions)
            position_tolerance = 10  # xy tol = tolerance in um
            first_frame_of_each_fov = [0]
            last_frame_of_each_fov = []
            for fri in range(n_frames - 1):
                if abs(x_positions[fri] - x_positions[fri + 1]) > position_tolerance or abs(
                        y_positions[fri] - y_positions[fri + 1]) > position_tolerance:
                    first_frame_of_each_fov.append(fri + 1)
                    last_frame_of_each_fov.append(fri)
            last_frame_of_each_fov.append(n_frames - 1)

            self.number_of_fov = len(first_frame_of_each_fov)
            self.first_frame_of_each_fov = first_frame_of_each_fov
            self.last_frame_of_each_fov = last_frame_of_each_fov


            self.width = self.file.metadata['width']
            self.height = self.file.metadata['height']
            self.pixel_size = np.array([self.file.metadata['pixel_microns'], self.file.metadata['pixel_microns']])

            # self.number_of_fields_of_view = len(images.metadata["experiment"]["loops"])  # number of fov is now available from self.fov_info
            self.number_of_frames = len(self.file)

            self.illuminations = [Illumination(name) for name in self.file.metadata["channels"]]

            if self.fov_index is not None:
                self.frame_offset = self.first_frame_of_each_fov[self.fov_index] * self.number_of_illuminations
                frame_end = (self.last_frame_of_each_fov[self.fov_index]+1) * self.number_of_illuminations
                self.number_of_frames = frame_end - self.frame_offset
            else:
                self.frame_offset = 0
                frame_end = self.number_of_frames

                self.stage_coordinates = np.array([[x_positions[0], y_positions[0]]])
                self.stage_coordinates_in_pixels = self.stage_coordinates / self.pixel_size


            self.illumination_arrangement = np.arange(len(self.illuminations))

            if self.fov_index is not None:
                self.time = xr.DataArray(np.repeat(
                    self.file.timesteps[self.first_frame_of_each_fov[self.fov_index]:(self.last_frame_of_each_fov[self.fov_index]+1)],
                    self.number_of_illuminations)/1000, dims='frame', coords={}, attrs={'units': 's'})
            else:
                self.time = xr.DataArray(np.repeat(self.file.timesteps, self.number_of_illuminations)/1000, dims='frame',
                                         coords={}, attrs={'units': 's'})

            # self.exp_time = images.metadata['experiment']['loops'][0]['sampling_interval']
            # self.exp_time_start=images.metadata['experiment']['loops'][0]['start']
            # self.exp_time_duration=images.metadata['experiment']['loops'][0]['duration']
            # self.pixelmicron=images.metadata['experiment']['pixel_microns']



    def _read_frame(self, frame_number):
        with self:
            if frame_number < self.number_of_frames:
                im = self.file[frame_number + self.frame_offset]
            else:
                im = self.file[self.number_of_frames + self.frame_offset - 1]
                print(f'pageNb out of range. The last frame (fr#{self.number_of_frames + self.frame_offset - 1}) is loaded instead')
            # note: im is a Frame, which is pims.frame.Frame, a np. array with additional frame number and metadata
            return im

    def _read_frames(self, indices):
        # Can probably be implemented more efficiently
        return np.stack([self._read_frame(i) for i in indices])

#
# def get_fov_from_nd2(nd2_fullpath):
#     images = ND2Reader(str(nd2_fullpath))
#     y_positions = images._parser._raw_metadata.y_data  # nikon sample stage position
#     x_positions = images._parser._raw_metadata.x_data  # nikon sample stage position
#
#     # set the image data order in the nd2 file
#     if 'c' in images.axes:
#         images.iter_axes = 'tc'  # for alex measurements
#     else:
#         images.iter_axes = 't'
#
#     n_illumination = len(images.metadata["channels"])
#     n_frames = len(x_positions)
#     position_tolerance = 10  # xy tol = tolerance in um
#     first_frame_of_each_fov = [0]
#     last_frame_of_each_fov = []
#     for fri in range(n_frames - 1):
#         if abs(x_positions[fri] - x_positions[fri + 1]) > position_tolerance or abs(
#                 y_positions[fri] - y_positions[fri + 1]) > position_tolerance:
#             first_frame_of_each_fov.append(fri + 1)
#             last_frame_of_each_fov.append(fri)
#     last_frame_of_each_fov.append(n_frames - 1)
#     fov_info = {'number_of_fov': len(first_frame_of_each_fov),
#                 'first_frame_of_each_fov': first_frame_of_each_fov,
#                 'last_frame_of_each_fov': last_frame_of_each_fov}
#     return fov_info

if __name__ == "__main__":
    print('test')



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 16 16:59:54 2018

@author: Ivo Severins
"""

import numpy as np
#import pandas as pd
import matplotlib.pyplot as plt
import re
import gc
import copy
from tabulate import tabulate
from pathlib import Path
import pandas as pd
try:
    import logomaker
except ModuleNotFoundError:
    pass

import matplotlib.path as pth

class FastqData:
    def __init__(self, path):
        self.path = Path(path)
        self.write_path = self.path.parent
        self.adapter_sequences = ['AATGATACGGCGACCACCGAGATCTACACTCTTTCCCTACACGACGCTCTTCCGATCT', 'ATCTCGTATGCCGTCTTCTGCTTG']
        self._text_file = None
        self._tiles = []
        self.is_reverse_complement = False

        file = open(path, 'r')

        lineCount = 0
        recordCount = 0
        
        instrument = list()
        run = list()
        flowcell = list()
        lane = list()
        tile = list()
        x = list()
        y = list()
        sample = list()

        sequence = list()
        quality = list()
        
        
        for line in file:
            if lineCount % 4 == 0: 
                splitHeader = re.split('[@: -]',line.strip())
                
                instrument.append(splitHeader[1])
                run.append(splitHeader[2])
                flowcell.append(splitHeader[4])
                lane.append(int(splitHeader[5]))
                tile.append(int(splitHeader[6]))
                x.append(int(splitHeader[7]))
                y.append(int(splitHeader[8]))
                sample.append(int(splitHeader[12]))
                
            if lineCount % 4 == 1: 
                sequence.append(line.strip())
            if lineCount % 4 == 3: 
                quality.append(line.strip())
                       
                recordCount +=1
        
            lineCount +=1
        
        
        file.close()

        self.name = np.array(['']*len(sequence))
        self.instrument = np.array(instrument)
        self.run = np.array(run)
        self.flowcell = np.array(flowcell)
        self.lane = np.array(lane)
        self.tile = np.array(tile)
        self.x = np.array(x)
        self.y = np.array(y)
        self.sample = np.array(sample)
        self.sequence = np.array(sequence, dtype = bytes).view('S1').reshape((len(sequence),-1))
        self.quality = np.array(quality, dtype = bytes).view('S1').reshape((len(quality),-1))

        self.selection

        self.write_to_text_file('Total number of sequences: ' + str(self.sequence.shape[0]) + '\n\n')

    @property
    def tile_numbers(self):
        return np.unique(self.tile)

    @property
    def tiles(self):
        if not self._tiles:
            self._tiles = [self.get_tile_object(tile_number) for tile_number in self.tile_numbers]
        return self._tiles

    @property
    def coordinates(self):
        return np.vstack([self.x, self.y]).T

    @property
    def text_file(self):
        if not self._text_file:
            self._text_file = self.write_path.joinpath('Output.txt')
            with self._text_file.open('w') as f:
                f.write('Analysis\n\n')
        return self._text_file

    def write_to_text_file(self, input_text):
        with self.text_file.open('a') as f:
            f.write(input_text + '\n')

    def export_fastq(self, filepath = None):
        if not filepath: filepath = self.write_path

        with filepath.with_suffix('.fastq').open('w') as f:
            for i in np.arange(len(self)):
                f.write(f"@{self.instrument[i]}:{self.run[i]}:000000000-{self.flowcell[i]}:{self.lane[i]}:"
                        f"{self.tile[i]}:{self.x[i]}:{self.y[i]} 1:N:0:{self.sample[i]}\n"
                        f"{self.sequence[i].tostring().decode('utf-8')}\n"
                        f"+\n"
                        f"{self.quality[i].tostring().decode('utf-8')}\n"
                        )

    def __len__(self):
        return self.sequence.shape[0]

    def __getitem__(self, item):
        new = copy.copy(self)
        new.name = self.name[item]
        new.instrument = self.instrument[item]
        new.run = self.run[item]
        new.flowcell = self.flowcell[item]
        new.lane = self.lane[item]
        new.tile = self.tile[item]
        new.x = self.x[item]
        new.y = self.y[item]
        new.sample = self.sample[item]
        new.sequence = self.sequence[item, :]
        new.quality = self.quality[item, :]

        new._tiles = []
        return new

    def __add__(self, other):
        new = copy.copy(self)
        new.name = np.append(new.name, other.name)
        new.instrument = np.append(new.instrument, other.instrument)
        new.run = np.append(new.run, other.run)
        new.flowcell = np.append(new.flowcell, other.flowcell)
        new.lane = np.append(new.lane, other.lane)
        new.tile = np.append(new.tile, other.tile)
        new.x = np.append(new.x, other.x)
        new.y = np.append(new.y, other.y)
        new.sample = np.append(new.sample, other.sample)
        new.sequence = np.vstack([new.sequence, other.sequence])
        new.quality = np.vstack([new.quality, other.quality])

        new._tiles = []
        return new

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def select(self, indices, copyData=False):
        if copyData:
            selection = copy.deepcopy(self)
        else:
            selection = self

        selection.name = self.name[indices]
        selection.instrument = self.instrument[indices]
        selection.run = self.run[indices]
        selection.flowcell = self.flowcell[indices]
        selection.lane = self.lane[indices]
        selection.tile = self.tile[indices]
        selection.x = self.x[indices]
        selection.y = self.y[indices]
        selection.sample = self.sample[indices]
        selection.sequence = self.sequence[indices,:]
        selection.quality = self.quality[indices,:]

        selection._tiles = []
        
        if copyData:
            return selection

    def selection(self, save=False, **kwargs):
        selection = np.zeros((self.sequence.shape[0], len(kwargs)), dtype=bool)
        for i, (key, value) in enumerate(kwargs.items()):
            if key == 'sequence':
                selection[:, i] = np.all(self.sequence[:, 0:len(value)] == np.array(list(value), dtype = bytes), axis = 1)
            elif key in ['x', 'y']:
                selection[:, i] = np.all(np.vstack([getattr(self, key) > np.min(value), getattr(self, key) < np.max(value)]), axis=0)
            elif key == 'coordinates_within_vertices':
                selection[:, i] = pth.Path(value).contains_points(self.coordinates)
            elif key == 'boolean_selection':
                selection[:, i] = value
            elif key == 'in_name':
                if not isinstance(value, list):
                    value = [value]
                selection[:, i] = [any(test_string in name for test_string in value)
#                                    if name is not None else False
                                   for name in self.name]
            else:
                selection[:, i] = getattr(self, key) == value

        return np.all(selection, axis=1)

    def get_selection(self, **kwargs):
        selection = self.selection(**kwargs)
        return self[selection]

    def reverse_complement(self):
        condlist = [self.sequence == b'A', self.sequence == b'T',
                    self.sequence == b'C', self.sequence == b'G']
        choicelist = [b'T', b'A', b'G', b'C']
        self.sequence = np.fliplr(np.select(condlist, choicelist))
        self.quality = np.fliplr(self.quality)
        self.is_reverse_complement = not self.is_reverse_complement

    def base_count(self):
        base_count = pd.DataFrame()
        for b in ['A', 'C', 'T', 'G']:
            base_count[b] = np.sum(self.sequence == b.encode(), axis=0)
        return base_count

    def logo_plot(self, start=None, end=None, row_length=None, figure=None, save=False, title='', **kwargs):
        base_count = self.base_count()[slice(start, end)]
        if not row_length:
            row_length=len(base_count)
        n_rows = len(base_count) // row_length
        if not figure:
            figure = plt.figure(figsize=(row_length/5,2*n_rows))

        figure.subplots(n_rows,1, sharey=True)
        axes = figure.axes

        for i in np.arange(n_rows):
            logo_object = logomaker.Logo(base_count[(i*row_length):((i+1)*row_length)], ax=axes[i], **kwargs)
            axes[i].set_ylabel('Count')

        axes[0].set_title(title)
        axes[-1].set_xlabel('Position')


        figure.tight_layout(pad=1)
        if save:
            title = title.replace('>','gt').replace('<','st')
            figure.savefig(f'{title}_seqlogo.png')

        return logo_object

    def sequence_density(self, expected_seq=None, start=None, end=None, row_length=None, figure=None, save=False, title=''):
        # TODO: Replace with function in plotting_sequences.py

        sequence = self.sequence[~np.any(self.sequence == b'N', axis=1)]
        # sequence_int = sequence.view('uint8')
        # sequence_df = pd.DataFrame(sequence_int).iloc[:, slice(start, end)]
        sequence_df = pd.DataFrame(sequence)
        expected_seq = expected_seq[slice(start, end)]

        total_length = sequence_df.shape[1]

        if not row_length:
            row_length = total_length

        n_rows = total_length // row_length
        if not figure:
            figure = plt.figure(figsize=(row_length/5,2*n_rows))

        figure.subplots(n_rows, 1, sharey=False)
        axes = figure.axes

        bases = [b'A', b'T', b'C',b'G']
        index = [(b1,b2) for b1 in bases for b2 in bases]
        # out = pd.DataFrame(index=index)
        for r in np.arange(n_rows):
            for i in np.arange((r*row_length),((r+1)*row_length)-1):
                print(i)
                # out[i] = (test.iloc[:,i:(i+2)].value_counts(normalize=True))
                base_transition_count = (sequence_df.iloc[:, i:(i + 2)].value_counts(normalize=True)).reindex(index)
                ys = base_transition_count.index.to_list()
                zs = base_transition_count.values
                x = (i, i + 1)
                for y, z in zip(ys, zs):
                    axes[r].plot(x, y, lw=z * 10, c='b', solid_capstyle='round')
                # ax.plot((i, i + 1), [0, 0])
            # expected_seq=sequences_per_read['Read1']['HJ1']
            if expected_seq:
                # axis.set_xticklabels(list(expected_seq))
                # axis.set_xticks(np.arange(len(expected_seq)))
                xs = np.where(np.array(list(expected_seq)) != 'N')[0]
                x0 = 0
                for s in re.split('N', expected_seq):
                    if s:
                        x = xs[x0:(x0 + len(s))]
                        y = np.array(list(s), dtype='S1')#.view('uint8')
                        axes[r].plot(x, y, c='r')
                        x0 += len(s)
            axes[r].set_xlim(r*row_length, (r+1)*row_length-1)
            axes[r].set_ylabel('Base')
            # axes[r].xaxis.set_minor_locator(plt.MultipleLocator(1))

        axes[0].set_title(title)
        axes[-1].set_xlabel('Position')

        # for axis in axes:
        #     bases = list('ATGC')
        #     axis.set_yticks(np.array(bases, dtype='S1').view('uint8'))
        #     axis.set_yticklabels(bases)

        figure.tight_layout(pad=1)
        if save:
            title = title.replace('>','gt').replace('<','st')
            figure.savefig(f'{title}_seqdensity.png')

    def number_of_matches(self, sequence):
        # sequence must be a string
        # return np.sum(self.sequence[:,0:len(sequence)]==np.array(list(sequence), dtype = bytes),1)

        sequence_bytes = np.array(list(sequence), dtype=bytes)
        indices_not_N = np.where(sequence_bytes != np.array('N', dtype=bytes))[0]
        number_of_Ns = len(sequence)-len(indices_not_N)

        return np.sum(self.sequence[:, indices_not_N] == sequence_bytes[indices_not_N], 1) + number_of_Ns

    def number_of_matches_with_sequence_dict(self, sequence_dict):
        number_of_matches = [pd.Series(self.number_of_matches(sequence), name=sequence_name) for
                             sequence_name, sequence in sequence_dict.items()]
        number_of_matches = (pd.concat(number_of_matches, keys=sequence_dict.keys(), names='Sequence', axis=1)
                             .rename_axis('Sequence index').astype(pd.Int64Dtype()))
        return number_of_matches

    def number_of_mismatches_with_sequence_dict(self, sequence_dict):
        number_of_matches = self.number_of_matches_with_sequence_dict(sequence_dict)
        number_of_mismatches = number_of_matches.copy()
        for sequence in number_of_matches.columns:
            number_of_mismatches[sequence] = len(sequence_dict[sequence]) - number_of_matches[sequence]
        return number_of_mismatches

    def matches_per_tile(self, sequence):
        if len(sequence) < 5: return

        header = [''] + list(self.tile_numbers)
        emptyRow = ['' for i in np.arange(9)]

        number_of_matches = self.number_of_matches(sequence)

        allClusters = ['All clusters'] + [np.sum(self.selection(tile = tile)) for tile in self.tile_numbers]
        fullMatchCount = ['Full matches'] + [np.sum(self.tile[number_of_matches == (len(sequence))] == tile) for tile in self.tile_numbers]
        oneMisMatch = ['1 mismatch'] + [np.sum(self.tile[number_of_matches == (len(sequence) - 1)] == tile) for tile in self.tile_numbers]
        twoMisMatches = ['2 mismatches'] + [np.sum(self.tile[number_of_matches == (len(sequence) - 2)] == tile) for tile in self.tile_numbers]
        threeMisMatches = ['3 mismatches'] + [np.sum(self.tile[number_of_matches == (len(sequence) - 3)] == tile) for tile in self.tile_numbers]
        fourMisMatches = ['4 mismatches'] + [np.sum(self.tile[number_of_matches == (len(sequence) - 4)] == tile) for tile in self.tile_numbers]
        fiveMisMatches = ['5 mismatches'] + [np.sum(self.tile[number_of_matches == (len(sequence) - 5)] == tile) for tile in self.tile_numbers]

        # lessThan3mismatches = ['<=2 mismatches'] + [np.sum(data.tile[Nmatch > 47] == tile) for tile in self.tile_numbers]]

        table = tabulate([allClusters,
                          emptyRow,
                          fullMatchCount,
                          oneMisMatch,
                          twoMisMatches,
                          threeMisMatches,
                          fourMisMatches,
                          fiveMisMatches,
                          emptyRow
                          ], header)

        print(table)
        self.write_to_text_file('Matches per tile with sequence: ' + sequence + '\n\n')
        self.write_to_text_file(table)

    def classify(self, sequences_dict, criteria_dict):
        df = self.number_of_mismatches_with_sequence_dict(sequences_dict)

        selection = pd.DataFrame(index=df.index)
        for sequence_name, criterion in criteria_dict.items():
            selection.loc[df.query(criterion, engine='python').index, sequence_name] = True

            self.name = selection.apply(lambda x: ', '.join(x.index[x == True]), axis=1).to_numpy()

    def get_tile_object(self, tile):
        x = self.x[self.selection(tile=tile)]
        y = self.y[self.selection(tile=tile)]
        return Tile(tile, np.transpose([x,y]))
        # tile = Tile('2102', np.loadtxt(path.joinpath('2102.loc'))) If we later want to get it from the file

    def export_positions_per_tile(self):
        # Export clusterpositions
        for tile in self.tile_numbers:
            x = self.x[self.selection(tile=tile)]
            y = self.y[self.selection(tile=tile)]
            np.savetxt(self.write_path.joinpath(str(tile)+'.loc'), np.transpose([x,y]), fmt='%u', delimiter='\t')

    def show_tiles(self):
        # Figure with matching cluster positions
        fig, axes = plt.subplots(2, 4, figsize=(16, 8), sharex='all', sharey='all')
        for i, tile in enumerate(self.tile_numbers):
            ax = axes.flatten()[i]
            x = self.x[self.selection(tile = tile)]
            y = self.y[self.selection(tile = tile)]
            ax.scatter(x, y, c='k', marker='.')
            ax.set_title('Tile ' + str(tile))
            ax.set_aspect('equal')

            ax.set_xlim([0, 31000])
            ax.set_ylim([0, 31000])

            if i in [4, 5, 6, 7]: ax.set_xlabel('x (FASTQ)')
            if i in [0, 4]: ax.set_ylabel('y (FASTQ)')

        fig.savefig(self.write_path.joinpath('clusterPositionsWithMatchingSequence.pdf'), bbox_inches='tight')
        fig.savefig(self.write_path.joinpath('clusterPositionsWithMatchingSequence.png'), bbox_inches='tight')


class Tile:
    def __init__(self, number, coordinates):
        self.name = str(number)
        self.number = number
        self.coordinates = coordinates

    def __repr__(self):
        return (f'{self.__class__.__name__}({self.name})')

#https://stackoverflow.com/questions/9476797/how-do-i-create-character-arrays-in-numpy

#dataDictionary = {
#		'instrument': instrument,
#		'run': run,
#		'flowcell': flowcell,
#		'lane': lane,
#		'tile': tile,
#		'x': x,
#		'y': y,
#		'sample': sample,
#		'sequence': sequence,
#		'quality': quality
#		}
#
#
#df = pd.DataFrame(columns=['instrument','run','flowcell','lane','tile','x','y','sample','sequence','quality'])
#df = pd.DataFrame(dataDictionary)

# A way to get out specific sequences
#regex = re.compile('(\w){0}AA(\w){49}')
#len(list(filter(regex.match, sequence)))


#
#
#[df['sequence'][:][i] == df['sequence'][1][i] for i in range(len(df['sequence'][1]))]
#
#gc.collect()
#
#
#seqList = list(map(list, sequence))
#seqArray = np.array(list(map(np.array,seqList)))
#
#compSeq = np.array(list('ACTGTTTTTTTTTTTTTTTTACTACCTCTTTTTTTTTTTTTTT'))
#df['mmCountCompSeq'] = np.sum(seqArray[:,0:43]==compSeq, axis=1)
#df['firstMmPosition'] = np.logical_not((seqArray[:,0:43]==compSeq)).argmax(axis=1)
#
#
#df[df['mmCountCompSeq']==42]['firstMmPosition'].plot.hist(bins = 43)


# =============================================================================
# def matchCount(row):
# 	return sum([x==y for x,y in zip(row['sequence'],'CCTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTCTTTTTC')])
# 
# 
# df.apply(lambda row: matchCount(row), axis=1)
# =============================================================================

# =============================================================================
# 
# 
# def matchCount(row):
#	return sum([x==y for x,y in zip(row['sequence'],'CCTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTCTTTTTC')])
# test = list()
# for seq in sequence:
# 	test.append(sum([x==y for x,y in zip(seq,'GATTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTCTTTTTC')]))
# 	
# 	
# 		
# test = list()
# for seq in sequence:
# 	test.append([ord(char) for char in seq])
# 	
# a = np.array(test)
# 
# b = np.sum(a[1,:]==a,axis=1)
# =============================================================================

# =============================================================================
# df.sequence.str[1:3]
# 
# 
# 
# a = np.array([list(x) for x in df.sequence[1:3].str[1:40]])
# np.sum(df.sequence.str[1:3]==['T','T'],axis=1)
# =============================================================================





# =============================================================================
# 
# from Bio import SeqIO
# #record_dict = SeqIO.to_dict(SeqIO.parse(r"C:\Users\Ivo Severins\Desktop\Sequencing data\20180705\Undetermined_S0_L001_I1_001.fastq","fastq"))
# 

# 
# 
# 
# records = list(SeqIO.parse(path+file, "fastq"))
# 
# x = [o.id[33:38] for o in records]
# 
# 
# x = [re.search('(?=:)...',o.id) for o in records]
# 
# re.split(
# 
# 
# test = [o.letter_annotations['phred_quality'] for o in records]
# 
# 
# 
# 
# 
# 
# file = open(path+file, 'r') 
# print file.readline(): 
# =============================================================================









import sys
from pathlib import Path
sys.path.append(str(Path(__file__).parents[1]))

from papylio.experiment import Experiment
from papylio.file import File
# from papylio.molecule import Molecule
from papylio.movie.movie import Movie
from papylio.analysis.interactiveAnalysis import InteractivePlot

# import papylio.sequencing.sequencing
#
# Experiment = type(Experiment.__name__, (Experiment,) + tuple(Experiment.plugins), {})

try:
    from ._version import version as __version__
except ImportError:
    try:
        import setuptools_scm
        __version__ = setuptools_scm.get_version(version_scheme="post-release")
    except (LookupError, ImportError):
        __version__ = '0.0.0'


import time

class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""

class Timer:
    def __init__(self):
        self._start_time = None
        self.name = None

    def start(self, name=''):
        """Start a new timer"""
        if self._start_time is not None:
            raise TimerError(f"Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter()
        self.name = name

    def stop(self):
        """Stop the timer, and report the elapsed time"""
        if self._start_time is None:
            raise TimerError(f"Timer is not running. Use .start() to start it")

        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None

        print(f"{self.name} \t Elapsed time: {elapsed_time:0.4f} seconds")
        self.name = None

    def startstop(self, name=''):
        self.stop()
        self.start(name=name)


import matplotlib.pyplot as plt
import pandas as pd
import xarray as xr
import numpy as np
import itertools
import tqdm
from matplotlib import patches
from papylio.plugins.holliday_junction.sequence_generation import all_basepaired_subsets


def basepair_count_per_position(basepair_count, save_path):
    title = (basepair_count.name.replace('basepair_count','basepair_count_per_position'))

    basepair_count_stacked = basepair_count.stack(basepair=('base_0', 'base_1'))
    basepair_count_stacked['basepair'] = xr.concat([basepair_count_stacked.base_0, basepair_count_stacked.base_1],
                                                   dim='dummy').str.join(dim='dummy').values

    basepaired_position = xr.concat(
        [basepair_count_stacked.sel(position_0=p0, position_1=p1) for p0, p1 in [[7, 0], [1, 2], [3, 4], [5, 6]]],
        dim='position')

    cm = plt.colormaps['tab20c']
    figure, axis = plt.subplots(figsize=(10,3), layout='constrained')
    axis.cla()
    x = np.arange(len(basepaired_position.position)) * (len(basepaired_position.basepair) + 4)
    labels = []
    x_ticks = []
    for i, basepair in enumerate(basepaired_position.basepair.values):
        axis.bar(x + i, basepaired_position.sel(basepair=basepair), width=1, label=basepair, color=cm.colors[i])
        labels += [basepair[0] + '\n' + basepair[1]] * len(x)
        x_ticks += (x+i).tolist()
    # ax.legend()
    axis.set_xticks(x_ticks, labels)
    axis.set_ylabel('Count')
    axis.set_title(title)

    axis.annotate('Basepair 8-1', xy=(8,-40), xycoords=('data','axes points'), ha='center')
    axis.annotate('Basepair 2-3', xy=(28,-40), xycoords=('data','axes points'), ha='center')
    axis.annotate('Basepair 4-5', xy=(48,-40), xycoords=('data','axes points'), ha='center')
    axis.annotate('Basepair 6-7', xy=(68,-40), xycoords=('data','axes points'), ha='center')
    savefile_path = save_path / title
    figure.savefig(savefile_path.with_suffix('.png'))
    figure.savefig(savefile_path.with_suffix('.pdf'))

def basepaired_sequence_subset_count(sequence_subset_count, save_path):
    title = (sequence_subset_count.name.replace('sequence_count', 'basepaired_sequence_count'))
    variable = sequence_subset_count.attrs['variable']

    sequence_subset_count2 = sequence_subset_count.reindex(**{variable: all_basepaired_subsets(), 'fill_value': 0})

    # bases = ['A', 'T', 'C', 'G']
    basepairs = ['AT', 'TA', 'CG', 'GC']
    # possible_sequence_subsets = [''.join(ss) for ss in
    #                              itertools.product(bases, basepairs, bases, bases, basepairs, bases)]
    # possible_sequence_subset_count = sequence_subset_count.reindex(sequence_subset=possible_sequence_subsets,
    #                                                                fill_value=0)

    sequence_subset_count_bp = xr.DataArray(0, dims=('bp0', 'bp1', 'bp2', 'bp3'),
                                            coords={'bp0': basepairs, 'bp1': basepairs, 'bp2': basepairs,
                                                    'bp3': basepairs}).astype(sequence_subset_count2.dtype)
    for bpc in tqdm.tqdm(itertools.product(basepairs, basepairs, basepairs, basepairs)):
        sequence_subset_count_bp.loc[dict(bp0=bpc[0], bp1=bpc[1], bp2=bpc[2], bp3=bpc[3])] = \
            sequence_subset_count2.sel(**{variable: (''.join(bpc)[1:] + ''.join(bpc)[0])})

    figure, axis = plt.subplots(figsize=(7.2, 6.5))#, layout='tight')
    # axis = axes[0]
    axis.cla()
    data = sequence_subset_count_bp.stack(bp02=('bp0', 'bp2')).stack(bp13=('bp1', 'bp3')).T
    image = axis.imshow(data.values, vmin=0, vmax=sequence_subset_count_bp.mean() * 2, cmap='coolwarm')
    # axis.images[0].set_data(data.values)

    for x, bp in zip(np.arange(0, 16, 4) + 1.5, basepairs):
        axis.annotate(bp[::-1], xy=(x, 16), xycoords=('data', 'data'), ha='center', va='center')
    for y, bp in zip(np.arange(0, 16, 4) + 1.5, basepairs):
        axis.annotate(bp, xy=(-1, y), xycoords=('data', 'data'), ha='center', va='center')
    for x, bp in zip(np.arange(0, 16), basepairs * 4):
        axis.annotate(bp, xy=(x, -1), xycoords=('data', 'data'), ha='center', va='center')
    for y, bp in zip(np.arange(0, 16), basepairs * 4):
        axis.annotate(bp, xy=(16, y), xycoords=('data', 'data'), ha='center', va='center')

    axis.set_xticks(np.arange(-0.5, 16, 4)[1:-1], minor=False)
    # axis.set_xticks(np.arange(0,16), basepairs*4, minor=True)
    axis.set_yticks(np.arange(-0.5, 16, 4)[1:-1], minor=False)
    # axis.set_yticks(np.arange(0,16), basepairs*4, minor=True)
    axis.tick_params(which="major", top=False, labeltop=False, right=False, labelright=False,
                     bottom=False, labelbottom=False, left=False, labelleft=False)
    axis.grid(which='major', color="w", linestyle='-', linewidth=2)
    for spine in axis.spines.values():
        spine.set_visible(False)
    axis.annotate('Basepair 1-8', xy=(7.5, 17), xycoords=('data', 'data'), ha='center', va='center')
    axis.annotate('Basepair 2-3', xy=(-2, 7.5), xycoords=('data', 'data'), ha='center', va='center',
                  rotation='vertical')
    axis.annotate('Basepair 4-5', xy=(7.5, -2), xycoords=('data', 'data'), ha='center', va='center')
    axis.annotate('Basepair 6-7', xy=(17, 7.5), xycoords=('data', 'data'), ha='center', va='center',
                  rotation='vertical')

    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(axis)
    cax = divider.append_axes("right", "4%", pad="15%")
    figure.colorbar(image, cax=cax)

    # figure.colorbar(image, aspect=30)
    # cax = figure.axes[-1]
    cax.set_ylabel('Count')
    cax.axes.ticklabel_format(scilimits=(0,0))
    for spine in cax.spines.values():
        spine.set_visible(False)

    fontsize = 8
    title2 = title + ' - scaled_mean_basepaired'
    figure.suptitle(title2, fontsize=fontsize)
    savefile_path = save_path / title2
    figure.savefig(savefile_path.with_suffix('.png'))
    figure.savefig(savefile_path.with_suffix('.pdf'))

    image.set_clim(sequence_subset_count_bp.min(), np.percentile(sequence_subset_count.values, 20))
    title2 = title + ' - scaled_bottom_20%_of_all'
    figure.suptitle(title2, fontsize=fontsize)
    savefile_path = save_path / title2
    figure.savefig(savefile_path.with_suffix('.png'))
    figure.savefig(savefile_path.with_suffix('.pdf'))

    image.set_clim(sequence_subset_count_bp.min(), sequence_subset_count.mean().item() * 2)
    title2 = title + ' - scaled_mean_of_all'
    figure.suptitle(title2, fontsize=fontsize)
    savefile_path = save_path / title2
    figure.savefig(savefile_path.with_suffix('.png'))
    figure.savefig(savefile_path.with_suffix('.pdf'))

    image.set_clim(sequence_subset_count_bp.min(), sequence_subset_count_bp.max())
    title2 = title + ' - scaled_min_max_basepaired'
    figure.suptitle(title2, fontsize=fontsize)
    savefile_path = save_path / title2
    figure.savefig(savefile_path.with_suffix('.png'))
    figure.savefig(savefile_path.with_suffix('.pdf'))




# def plot_basepaired_holliday_junction(data, save_path, **imshow_kwargs):
#
#     data = data.reindex(**{'sequence_subset': all_basepaired_subsets(), 'fill_value': np.nan})
#
#     # bases = ['A', 'T', 'C', 'G']
#     basepairs = ['AT', 'TA', 'CG', 'GC']
#
#     data_bp = xr.DataArray(0, dims=('bp0', 'bp1', 'bp2', 'bp3'),
#                            coords={'bp0': basepairs, 'bp1': basepairs, 'bp2': basepairs,
#                                    'bp3': basepairs}).astype(data.dtype)
#     for bpc in tqdm.tqdm(itertools.product(basepairs, basepairs, basepairs, basepairs)):
#         data_bp.loc[dict(bp0=bpc[0], bp1=bpc[1], bp2=bpc[2], bp3=bpc[3])] = \
#             data.sel(**{'sequence_subset': (''.join(bpc)[1:] + ''.join(bpc)[0])})
#
#     figure, axis = plt.subplots(figsize=(7.9, 6.5), layout='constrained')
#     # axis = axes[0]
#     axis.cla()
#     data_bp_stacked = data_bp.stack(bp02=('bp0', 'bp2')).stack(bp13=('bp1', 'bp3')).T
#     image = axis.imshow(data_bp_stacked.values, cmap='coolwarm', **imshow_kwargs)
#     # axis.images[0].set_data(data.values)
#
#     for x, bp in zip(np.arange(0, 16, 4) + 1.5, basepairs):
#         axis.annotate(bp[::-1], xy=(x, 16), xycoords=('data', 'data'), ha='center', va='center')
#     for y, bp in zip(np.arange(0, 16, 4) + 1.5, basepairs):
#         axis.annotate(bp, xy=(-1, y), xycoords=('data', 'data'), ha='center', va='center')
#     for x, bp in zip(np.arange(0, 16), basepairs * 4):
#         axis.annotate(bp, xy=(x, -1), xycoords=('data', 'data'), ha='center', va='center')
#     for y, bp in zip(np.arange(0, 16), basepairs * 4):
#         axis.annotate(bp, xy=(16, y), xycoords=('data', 'data'), ha='center', va='center')
#
#     axis.set_xticks(np.arange(-0.5, 16, 4)[1:-1], minor=False)
#     # axis.set_xticks(np.arange(0,16), basepairs*4, minor=True)
#     axis.set_yticks(np.arange(-0.5, 16, 4)[1:-1], minor=False)
#     # axis.set_yticks(np.arange(0,16), basepairs*4, minor=True)
#     axis.tick_params(which="major", top=False, labeltop=False, right=False, labelright=False,
#                      bottom=False, labelbottom=False, left=False, labelleft=False)
#     axis.grid(which='major', color="w", linestyle='-', linewidth=2)
#     for spine in axis.spines.values():
#         spine.set_visible(False)
#     axis.annotate('Basepair 1-8', xy=(7.5, 17), xycoords=('data', 'data'), ha='center', va='center')
#     axis.annotate('Basepair 2-3', xy=(-2, 7.5), xycoords=('data', 'data'), ha='center', va='center',
#                   rotation='vertical')
#     axis.annotate('Basepair 4-5', xy=(7.5, -2), xycoords=('data', 'data'), ha='center', va='center')
#     axis.annotate('Basepair 6-7', xy=(17, 7.5), xycoords=('data', 'data'), ha='center', va='center',
#                   rotation='vertical')
#
#     from mpl_toolkits.axes_grid1 import make_axes_locatable
#     divider = make_axes_locatable(axis)
#     cax = divider.append_axes("right", "4%", pad="15%")
#     figure.colorbar(image, cax=cax)
#
#     # figure.colorbar(image, aspect=30)
#     # cax = figure.axes[-1]
#     cax.set_ylabel(data.name)
#     cax.axes.ticklabel_format(scilimits=(0,0))
#     for spine in cax.spines.values():
#         spine.set_visible(False)
#
#     title = 'basepaired_HJ_' + data.name
#
#     fontsize = 8
#     # title2 = title + ' - scaled_mean_basepaired'
#     figure.suptitle(title, fontsize=fontsize)
#     savefile_path = save_path / title
#     figure.savefig(savefile_path.with_suffix('.png'))
#     figure.savefig(savefile_path.with_suffix('.pdf'))
#
#     # image.set_clim(sequence_subset_count_bp.min(), np.percentile(sequence_subset_count.values, 20))
#     # title2 = title + ' - scaled_bottom_20%_of_all'
#     # figure.suptitle(title2, fontsize=fontsize)
#     # savefile_path = save_path / title2
#     # figure.savefig(savefile_path.with_suffix('.png'))
#     # figure.savefig(savefile_path.with_suffix('.pdf'))
#     #
#     # image.set_clim(sequence_subset_count_bp.min(), sequence_subset_count.mean().item() * 2)
#     # title2 = title + ' - scaled_mean_of_all'
#     # figure.suptitle(title2, fontsize=fontsize)
#     # savefile_path = save_path / title2
#     # figure.savefig(savefile_path.with_suffix('.png'))
#     # figure.savefig(savefile_path.with_suffix('.pdf'))
#     #
#     # image.set_clim(sequence_subset_count_bp.min(), sequence_subset_count_bp.max())
#     # title2 = title + ' - scaled_min_max_basepaired'
#     # figure.suptitle(title2, fontsize=fontsize)
#     # savefile_path = save_path / title2
#     # figure.savefig(savefile_path.with_suffix('.png'))
#     # figure.savefig(savefile_path.with_suffix('.pdf'))
#
#


def plot_holliday_junction(data, size=1, name=None, s2max=None, geometry='square',
                                      axis_facecolor="lightgrey", save_path=None, vmin=None, vmax=None, axis=None, cmap=None):

    if name is None:
        name = data.name

    if data.ndim == 1:
        data = data.expand_dims('loop_dim', 0)

    data = xr.Dataset(dict(data=data))
    if np.isscalar(size):
        data['size'] = xr.DataArray(size, coords=data.data.coords)
    elif isinstance(size, (list, tuple)):
        data['size'] = xr.DataArray(np.full_like(data.data, np.array(size)[None,:].T, dtype='object'), dims=data.data.dims)
    elif isinstance(size, xr.DataArray):
        if size.ndim == 1:
            size = size.expand_dims({data.data.dims[0]: data.data.shape[0]}, 0)
        data['size'] = size

    if isinstance(geometry, str):
        data['geometry'] = xr.DataArray(geometry, coords=data.data.coords)
    elif isinstance(geometry, (list, tuple)):
        data['geometry'] = xr.DataArray(np.full_like(data.data, np.array(geometry)[None,:].T, dtype='object'), dims=data.data.dims)
    elif isinstance(geometry, xr.DataArray):
        if geometry.ndim == 1:
            geometry = geometry.expand_dims({data.data.dims[0]: data.data.shape[0]}, 0)
        data['geometry'] = geometry

    # data = data.reindex(**{'sequence_subset': all_basepaired_subsets(), 'fill_value': np.nan})
    data['sequence_subset'] = data['sequence_subset'].astype('U')

    # bases = ['A', 'T', 'C', 'G']
    basepairs = ['AT', 'TA', 'CG', 'GC']

    bp0 = np.char.add(data.sequence_subset.str[7:8].values, data.sequence_subset.str[0:1].values)
    bp1 = data.sequence_subset.str[1:3].values
    bp2 = data.sequence_subset.str[3:5].values
    bp3 = data.sequence_subset.str[5:7].values
    data_bp = data.assign_coords(bp=('sequence_subset',pd.MultiIndex.from_arrays([bp0, bp1, bp2, bp3], names=['bp0','bp1','bp2','bp3'])))
    data_bp = data_bp.swap_dims(sequence_subset='bp').unstack('bp')
    # data_bp = data_bp.sel(bp0=basepairs, bp1=basepairs, bp2=basepairs, bp3=basepairs)


    # data_bp = xr.DataArray(0, dims=('bp0', 'bp1', 'bp2', 'bp3'),
    #                        coords={'bp0': basepairs, 'bp1': basepairs, 'bp2': basepairs,
    #                                'bp3': basepairs}).astype(data.dtype)
    # for bpc in tqdm.tqdm(itertools.product(basepairs, basepairs, basepairs, basepairs)):
    #     data_bp.loc[dict(bp0=bpc[0], bp1=bpc[1], bp2=bpc[2], bp3=bpc[3])] = \
    #         data.sel(**{'sequence_subset': (''.join(bpc)[1:] + ''.join(bpc)[0])})
    if axis is None:
        figure, axis = plt.subplots(figsize=(7.9, 6.5), layout='constrained')
    else:
        figure = axis.figure
    # axis = axes[0]
    axis.cla()
    data_bp_stacked = data_bp.stack(bp13=('bp1', 'bp3')).stack(bp02=('bp0', 'bp2'))
    data_bp_stacked = data_bp_stacked.sel(bp02=~data_bp_stacked.data.isnull().all('bp13').values.squeeze())
    data_bp_stacked = data_bp_stacked.sel(bp13=~data_bp_stacked.data.isnull().all('bp02').values.squeeze())
    # image = axis.imshow(data_bp_stacked.values, cmap='coolwarm', **imshow_kwargs)

    data_bp_stacked2 = data_bp_stacked.copy()
    data_bp_stacked2 = data_bp_stacked2.assign_coords(
        dict(bp02i=('bp02', np.arange(len(data_bp_stacked.bp02))), bp13i=('bp13', np.arange(len(data_bp_stacked.bp13)))))\
        .unstack(('bp13', 'bp02')).stack(bp0123=('bp0', 'bp1', 'bp2', 'bp3'))
    axis.cla()
    # axis.scatter(x=data_bp_stacked2.bp02i, y=data_bp_stacked2.bp13i, c=data_bp_stacked2.data.values, cmap='coolwarm',
    #              vmin=0, vmax=1, s=data_bp_stacked2.size_data.values, marker='s')

    x = data_bp_stacked2.bp02i.values
    y = data_bp_stacked2.bp13i.values

    selection = (~np.isnan(x)) & (~np.isnan(y))
    x = x[selection]
    y = y[selection]
    from matplotlib.colors import Normalize
    if vmin is None:
        vmin = 0
    if vmax is None:
        vmax = data_bp_stacked2.data.max().item()
    norm = Normalize(vmin=vmin, vmax=vmax)
    # cmap = 'coolwarm'
    if cmap is None:
        cmap = plt.get_cmap('coolwarm').copy()
        cmap.set_bad([1, 1, 1, 1])
    from matplotlib.cm import ScalarMappable
    scalar_mappable = ScalarMappable(norm=norm, cmap=cmap)
    # c = plt.get_cmap('coolwarm')(norm(data_bp_stacked2.data.values))

    for i in range(data_bp_stacked2.data.shape[0]):
        c = scalar_mappable.to_rgba(data_bp_stacked2.data[i].values)[selection]

        s = np.sqrt(data_bp_stacked2.size[i].values)[selection]
        if s2max is None:
            s2max = np.nanmax(s)**2
        s = np.minimum(s/np.sqrt(s2max),1) #s/s.max()

        g = data_bp_stacked2.geometry[i].values[selection]

        for xi, yi, ci, si, gi in zip(x,y,c,s,g):
            shape_kwargs = dict(facecolor=ci, linewidth=None)
            if gi == 'square':
                rect = plt.Rectangle([xi - si / 2, yi - si / 2], si, si, **shape_kwargs)
            elif gi == 'circle':
                rect = plt.Circle([xi, yi], radius=si/2, **shape_kwargs)
            elif gi == 'diamond':
                si2 = si/np.sqrt(2)
                rect = plt.Rectangle([xi, yi - si / 2], si2, si2, angle=45, **shape_kwargs)
            elif gi == 'triangle_top_left':
                rect = plt.Polygon([[xi - si / 2, yi - si / 2],
                                    [xi - si / 2, yi + si / 2],
                                    [xi + si / 2, yi - si / 2]], closed=True, **shape_kwargs)
            elif gi == 'triangle_bottom_right':
                rect = plt.Polygon([[xi + si / 2, yi + si / 2],
                                    [xi - si / 2, yi + si / 2],
                                    [xi + si / 2, yi - si / 2]], closed=True, **shape_kwargs)
            elif gi == 'semi_circle_top_left':
                rect = patches.Wedge([xi, yi], r=si / 2, theta1=-225, theta2=-45, **shape_kwargs)
            elif gi == 'semi_circle_bottom_right':
                rect = patches.Wedge([xi, yi], r=si / 2, theta1=-45, theta2=-225, **shape_kwargs)
            else:
                rect = None
            if rect is not None:
                axis.add_patch(rect)
    axis.set_aspect(1)
    half_shape_size = np.nanmax(s)/2
    axis.set_xlim(np.nanmin(x)-half_shape_size, np.nanmax(x)+half_shape_size)
    axis.set_ylim(np.nanmin(y)-half_shape_size, np.nanmax(y)+half_shape_size)
    # axis.autoscale_view()

    # axis.images[0].set_data(data.values)


    x_len = len(np.unique(data_bp_stacked2.bp02i))
    y_len = len(np.unique(data_bp_stacked2.bp13i))

    bp0_labels, bp0_labels_x = np.unique(data_bp_stacked.bp0, return_counts=True)
    bp0_edges = np.hstack([[0], bp0_labels_x.cumsum()])
    bp0_labels_x = bp0_labels_x.cumsum() - bp0_labels_x / 2 - 0.5
    for label, x in zip(bp0_labels, bp0_labels_x):
        axis.annotate(label, xy=(x, y_len), xycoords=('data', 'data'), ha='center', va='center')

    bp1_labels, bp1_labels_y = np.unique(data_bp_stacked.bp1, return_counts=True)
    bp1_edges = np.hstack([[0], bp1_labels_y.cumsum()])
    bp1_labels_y = bp1_labels_y.cumsum() - bp1_labels_y / 2 - 0.5
    for label, y in zip(bp1_labels, bp1_labels_y):
        axis.annotate(label, xy=(-1, y), xycoords=('data', 'data'), ha='center', va='center')

    bp2_labels = data_bp_stacked.bp2.values
    bp2_labels_x = np.arange(len(bp2_labels))
    for label, x in zip(bp2_labels, bp2_labels_x):
        axis.annotate(label, xy=(x, -1), xycoords=('data', 'data'), ha='center', va='center')

    bp3_labels = data_bp_stacked.bp3.values
    bp3_labels_y = np.arange(len(bp3_labels))
    for label, y in zip(bp3_labels, bp3_labels_y):
        axis.annotate(label, xy=(x_len, y), xycoords=('data', 'data'), ha='center', va='center')


    axis.set_xticks(bp0_edges-0.5, minor=False)
    axis.set_yticks(bp1_edges-0.5, minor=False)
    axis.tick_params(which="major", top=False, labeltop=False, right=False, labelright=False,
                     bottom=False, labelbottom=False, left=False, labelleft=False)
    axis.grid(which='major', color="w", linestyle='-', linewidth=2)
    # axis.grid(which='major', color="k", linestyle='-', linewidth=2)

    axis.set_xticks(bp2_labels_x, minor=True)
    axis.set_yticks(bp3_labels_y, minor=True)
    axis.tick_params(which="minor", top=False, labeltop=False, right=False, labelright=False,
                     bottom=False, labelbottom=False, left=False, labelleft=False)
    axis.grid(which='minor', color="k", linestyle='-', linewidth=0)

    #
    # x_len = len(np.unique(data_bp_stacked2.bp02i))
    # x_interval = len(np.unique(data_bp_stacked2.bp2))
    #
    # y_len = len(np.unique(data_bp_stacked2.bp13i))
    # y_interval = len(np.unique(data_bp_stacked2.bp3))
    #
    # for i in np.arange(0, x_len, x_interval):
    #     x = i + x_interval/2-0.5
    #     bp = data_bp_stacked2.bp0[i].item()
    #     axis.annotate(bp, xy=(x, y_len), xycoords=('data', 'data'), ha='center', va='center')
    # for i in np.arange(0, y_len, y_interval):
    #     y = i + y_interval/2-0.5
    #     bp = data_bp_stacked2.bp1[i].item()
    #     axis.annotate(bp, xy=(-1, y), xycoords=('data', 'data'), ha='center', va='center')
    # for i in np.arange(0, x_len):
    #     x = i
    #     bp = data_bp_stacked2.bp2[i].item()
    #     axis.annotate(bp, xy=(x, -1), xycoords=('data', 'data'), ha='center', va='center')
    # for i in np.arange(0, y_len):
    #     y = i
    #     bp = data_bp_stacked2.bp3[i].item()
    #     axis.annotate(bp, xy=(x_len, y), xycoords=('data', 'data'), ha='center', va='center')
    #
    # axis.set_xticks(np.arange(-0.5, x_len, x_interval)[1:-1], minor=False)
    # axis.set_yticks(np.arange(-0.5, y_len, y_interval)[1:-1], minor=False)
    # axis.tick_params(which="major", top=False, labeltop=False, right=False, labelright=False,
    #                  bottom=False, labelbottom=False, left=False, labelleft=False)
    # axis.grid(which='major', color="w", linestyle='-', linewidth=2)
    # # axis.grid(which='major', color="k", linestyle='-', linewidth=2)
    #
    # axis.set_xticks(np.arange(-0.5,x_len + 0.5), minor=True)
    # axis.set_yticks(np.arange(-0.5,y_len + 0.5), minor=True)
    # axis.tick_params(which="minor", top=False, labeltop=False, right=False, labelright=False,
    #                  bottom=False, labelbottom=False, left=False, labelleft=False)
    # axis.grid(which='minor', color="k", linestyle='-', linewidth=0)

    for spine in axis.spines.values():
        spine.set_visible(False)
        spine.set_lw(2)

    axis.annotate('Base pair 7-8', xy=(x_len/2-0.5, y_len+1), xycoords=('data', 'data'), ha='center', va='center')
    axis.annotate('Base pair 1-2', xy=(-2, y_len/2-0.5), xycoords=('data', 'data'), ha='center', va='center',
                  rotation='vertical')
    axis.annotate('Base pair 3-4', xy=(x_len/2-0.5, -2), xycoords=('data', 'data'), ha='center', va='center')
    axis.annotate('Base pair 5-6', xy=(x_len+1, y_len/2-0.5), xycoords=('data', 'data'), ha='center', va='center',
                  rotation=270)

    axis.set_facecolor(axis_facecolor)
    axis.invert_yaxis()

    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(axis)
    bbox = axis.get_window_extent()
    width, height = bbox.width, bbox.height
    s = 16 / width
    p = 60 / width
    cax = divider.append_axes("right", f"{s*100}%", pad=f"{p*100}%")
    figure.colorbar(scalar_mappable, cax=cax)

    # figure.colorbar(image, aspect=30)
    # cax = figure.axes[-1]
    if 'unit' in data.data.attrs and data.data.attrs['unit'] !='':
        unit_string = f' ({data.data.attrs["unit"]})'
    else:
        unit_string = ''
    cax.set_ylabel(name + unit_string)
    # cax.axes.ticklabel_format(scilimits=(0,0))
    for spine in cax.spines.values():
        spine.set_visible(False)

    title = 'HJ_' + name

    fontsize = 8
    # title2 = title + ' - scaled_mean_basepaired'
    figure.suptitle(title, fontsize=fontsize)
    if save_path is not None:
        savefile_path = save_path / title
        figure.savefig(savefile_path.with_suffix('.png'))
        figure.savefig(savefile_path.with_suffix('.pdf'))



def plot_basepaired_holliday_junction(data, size=1, name=None, s2max=None, geometry='square',
                                      axis_facecolor="lightgrey", save_path=None, vmin=None, vmax=None, axis=None, cmap=None):

    if name is None:
        name = data.name

    if data.ndim == 1:
        data = data.expand_dims('loop_dim', 0)

    data = xr.Dataset(dict(data=data))
    if np.isscalar(size):
        data['size'] = xr.DataArray(size, coords=data.data.coords)
    elif isinstance(size, (list, tuple)):
        data['size'] = xr.DataArray(np.full_like(data.data, np.array(size)[None,:].T, dtype='object'), dims=data.data.dims)
    elif isinstance(size, xr.DataArray):
        if size.ndim == 1:
            size = size.expand_dims({data.data.dims[0]: data.data.shape[0]}, 0)
        data['size'] = size

    if isinstance(geometry, str):
        data['geometry'] = xr.DataArray(geometry, coords=data.data.coords)
    elif isinstance(geometry, (list, tuple)):
        data['geometry'] = xr.DataArray(np.full_like(data.data, np.array(geometry)[None,:].T, dtype='object'), dims=data.data.dims)
    elif isinstance(geometry, xr.DataArray):
        if geometry.ndim == 1:
            geometry = geometry.expand_dims({data.data.dims[0]: data.data.shape[0]}, 0)
        data['geometry'] = geometry

    data = data.reindex(**{'sequence_subset': all_basepaired_subsets(), 'fill_value': np.nan})
    data['sequence_subset'] = data['sequence_subset'].astype('U')

    # bases = ['A', 'T', 'C', 'G']
    basepairs = ['AT', 'TA', 'CG', 'GC']

    bp0 = np.char.add(data.sequence_subset.str[7:8].values, data.sequence_subset.str[0:1].values)
    bp1 = data.sequence_subset.str[1:3].values
    bp2 = data.sequence_subset.str[3:5].values
    bp3 = data.sequence_subset.str[5:7].values
    data_bp = data.assign_coords(bp=('sequence_subset',pd.MultiIndex.from_arrays([bp0, bp1, bp2, bp3], names=['bp0','bp1','bp2','bp3'])))
    data_bp = data_bp.swap_dims(sequence_subset='bp').unstack('bp')
    data_bp = data_bp.sel(bp0=basepairs, bp1=basepairs, bp2=basepairs, bp3=basepairs)


    # data_bp = xr.DataArray(0, dims=('bp0', 'bp1', 'bp2', 'bp3'),
    #                        coords={'bp0': basepairs, 'bp1': basepairs, 'bp2': basepairs,
    #                                'bp3': basepairs}).astype(data.dtype)
    # for bpc in tqdm.tqdm(itertools.product(basepairs, basepairs, basepairs, basepairs)):
    #     data_bp.loc[dict(bp0=bpc[0], bp1=bpc[1], bp2=bpc[2], bp3=bpc[3])] = \
    #         data.sel(**{'sequence_subset': (''.join(bpc)[1:] + ''.join(bpc)[0])})
    if axis is None:
        figure, axis = plt.subplots(figsize=(7.9, 6.5), layout='constrained')
    else:
        figure = axis.figure
    # axis = axes[0]
    axis.cla()
    data_bp_stacked = data_bp.stack(bp13=('bp1', 'bp3')).stack(bp02=('bp0', 'bp2'))
    # image = axis.imshow(data_bp_stacked.values, cmap='coolwarm', **imshow_kwargs)

    data_bp_stacked2 = data_bp_stacked.copy()
    data_bp_stacked2 = data_bp_stacked2.assign_coords(dict(bp02i=('bp02', np.arange(16)), bp13i=('bp13', np.arange(16))))\
        .unstack(('bp13', 'bp02')).stack(bp0123=('bp0', 'bp1', 'bp2', 'bp3'))
    axis.cla()
    # axis.scatter(x=data_bp_stacked2.bp02i, y=data_bp_stacked2.bp13i, c=data_bp_stacked2.data.values, cmap='coolwarm',
    #              vmin=0, vmax=1, s=data_bp_stacked2.size_data.values, marker='s')

    x = data_bp_stacked2.bp02i.values
    y = data_bp_stacked2.bp13i.values
    from matplotlib.colors import Normalize
    if vmin is None:
        vmin = 0
    if vmax is None:
        vmax = data_bp_stacked2.data.max().item()
    norm = Normalize(vmin=vmin, vmax=vmax)
    # cmap = 'coolwarm'
    if cmap is None:
        cmap = plt.get_cmap('coolwarm').copy()
        cmap.set_bad([1, 1, 1, 1])
    from matplotlib.cm import ScalarMappable
    scalar_mappable = ScalarMappable(norm=norm, cmap=cmap)
    # c = plt.get_cmap('coolwarm')(norm(data_bp_stacked2.data.values))

    for i in range(data_bp_stacked2.data.shape[0]):
        c = scalar_mappable.to_rgba(data_bp_stacked2.data[i].values)

        s = np.sqrt(data_bp_stacked2.size[i].values)
        if s2max is None:
            s2max = np.nanmax(s)**2
        s = np.minimum(s/np.sqrt(s2max),1) #s/s.max()

        g = data_bp_stacked2.geometry[i].values

        for xi, yi, ci, si, gi in zip(x,y,c,s,g):
            shape_kwargs = dict(facecolor=ci, linewidth=None)
            if gi == 'square':
                rect = plt.Rectangle([xi - si / 2, yi - si / 2], si, si, **shape_kwargs)
            elif gi == 'circle':
                rect = plt.Circle([xi, yi], radius=si/2, **shape_kwargs)
            elif gi == 'diamond':
                si2 = si/np.sqrt(2)
                rect = plt.Rectangle([xi, yi - si / 2], si2, si2, angle=45, **shape_kwargs)
            elif gi == 'triangle_top_left':
                rect = plt.Polygon([[xi - si / 2, yi - si / 2],
                                    [xi - si / 2, yi + si / 2],
                                    [xi + si / 2, yi - si / 2]], closed=True, **shape_kwargs)
            elif gi == 'triangle_bottom_right':
                rect = plt.Polygon([[xi + si / 2, yi + si / 2],
                                    [xi - si / 2, yi + si / 2],
                                    [xi + si / 2, yi - si / 2]], closed=True, **shape_kwargs)
            elif gi == 'semi_circle_top_left':
                rect = patches.Wedge([xi, yi], r=si / 2, theta1=-225, theta2=-45, **shape_kwargs)
            elif gi == 'semi_circle_bottom_right':
                rect = patches.Wedge([xi, yi], r=si / 2, theta1=-45, theta2=-225, **shape_kwargs)

            axis.add_patch(rect)
    axis.set_aspect(1)
    half_shape_size = np.nanmax(s)/2
    axis.set_xlim(x.min()-half_shape_size, x.max()+half_shape_size)
    axis.set_ylim(y.min()-half_shape_size, y.max()+half_shape_size)
    # axis.autoscale_view()

    # axis.images[0].set_data(data.values)

    for x, bp in zip(np.arange(0, 16, 4) + 1.5, basepairs):
        # axis.annotate(bp[::-1], xy=(x, 16), xycoords=('data', 'data'), ha='center', va='center')
        axis.annotate(bp, xy=(x, 16), xycoords=('data', 'data'), ha='center', va='center')
    for y, bp in zip(np.arange(0, 16, 4) + 1.5, basepairs):
        axis.annotate(bp, xy=(-1, y), xycoords=('data', 'data'), ha='center', va='center')
    for x, bp in zip(np.arange(0, 16), basepairs * 4):
        axis.annotate(bp, xy=(x, -1), xycoords=('data', 'data'), ha='center', va='center')
    for y, bp in zip(np.arange(0, 16), basepairs * 4):
        axis.annotate(bp, xy=(16, y), xycoords=('data', 'data'), ha='center', va='center')

    axis.set_xticks(np.arange(-0.5, 16, 4)[1:-1], minor=False)
    axis.set_yticks(np.arange(-0.5, 16, 4)[1:-1], minor=False)
    axis.tick_params(which="major", top=False, labeltop=False, right=False, labelright=False,
                     bottom=False, labelbottom=False, left=False, labelleft=False)
    axis.grid(which='major', color="w", linestyle='-', linewidth=2)
    # axis.grid(which='major', color="k", linestyle='-', linewidth=2)

    axis.set_xticks(np.arange(-0.5,16.5), minor=True)
    axis.set_yticks(np.arange(-0.5,16.5), minor=True)
    axis.tick_params(which="minor", top=False, labeltop=False, right=False, labelright=False,
                     bottom=False, labelbottom=False, left=False, labelleft=False)
    axis.grid(which='minor', color="k", linestyle='-', linewidth=0)

    for spine in axis.spines.values():
        spine.set_visible(False)
        spine.set_lw(2)

    axis.annotate('Base pair 7-8', xy=(7.5, 17), xycoords=('data', 'data'), ha='center', va='center')
    axis.annotate('Base pair 1-2', xy=(-2, 7.5), xycoords=('data', 'data'), ha='center', va='center',
                  rotation='vertical')
    axis.annotate('Base pair 3-4', xy=(7.5, -2), xycoords=('data', 'data'), ha='center', va='center')
    axis.annotate('Base pair 5-6', xy=(17, 7.5), xycoords=('data', 'data'), ha='center', va='center',
                  rotation=270)

    axis.set_facecolor(axis_facecolor)
    axis.invert_yaxis()

    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(axis)
    cax = divider.append_axes("right", "4%", pad="15%")
    figure.colorbar(scalar_mappable, cax=cax)

    # figure.colorbar(image, aspect=30)
    # cax = figure.axes[-1]
    if 'unit' in data.data.attrs and data.data.attrs['unit'] !='':
        unit_string = f' ({data.data.attrs["unit"]})'
    else:
        unit_string = ''
    cax.set_ylabel(name + unit_string)
    # cax.axes.ticklabel_format(scilimits=(0,0))
    for spine in cax.spines.values():
        spine.set_visible(False)

    title = 'basepaired_HJ_' + name

    fontsize = 8
    # title2 = title + ' - scaled_mean_basepaired'
    figure.suptitle(title, fontsize=fontsize)
    if save_path is not None:
        savefile_path = save_path / title
        figure.savefig(savefile_path.with_suffix('.png'))
        figure.savefig(savefile_path.with_suffix('.pdf'))


def format_sequence_subset(sequence_subset):
    ss = sequence_subset
    return np.array(
    [f'  ||  ',
    f'  {ss[3]}{ss[4]}  ',
    f'-{ss[2]}  {ss[5]}-',
    f'-{ss[1]}  {ss[6]}-',
    f'  {ss[0]}{ss[7]}  ',
    f'  ||  ']
    )

def format_sequence_subsets(sequence_subsets):
    spacer = np.array(['  ']*6)
    sequence_subsets_formatted = np.array([format_sequence_subset(ss) for ss in sequence_subsets])
    spacers = np.array([spacer]*len(sequence_subsets))

    final = char_add([np.char.add(*s) for s in zip(sequence_subsets_formatted, spacers)])
    final = '\n'.join(final)
    return final

# From itertools recipes
def batched(iterable, n):
    "Batch data into tuples of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch

def print_sequence_subsets(sequence_subsets, rows=None):
    if isinstance(sequence_subsets, str):
        sequence_subsets = [sequence_subsets]
    if rows is not None:
        sequence_subsets = batched(sequence_subsets, rows)
    else:
        sequence_subsets = [sequence_subsets]
    for sequence_subsets_row in sequence_subsets:
        print(format_sequence_subsets(sequence_subsets_row), '\n\n')


def char_add(strings):
    for i, s in enumerate(strings):
        if i == 0:
            final = s
        else:
            final = np.char.add(final, s)
    return final


def shorthand_notation(sequences):
    return [s[0::2] for s in sequences]


def purine_pyrimidine_sequence(sequences):
    translation_table = str.maketrans('AGCT', 'RRYY')
    return [s.translate(translation_table) for s in sequences]


def roll(sequences, n):
    for i in range(n):
        sequences = [s[1:] + s[0] for s in sequences]
    return sequences


def roll_multiple(sequences, n):
    ss = [sequences]
    for i in range(n):
        sequences = roll(sequences, 1)
        ss.append(sequences)
    return ss


def purine_pyrimidine_classes():
    return {1: 'RRRR', 2: 'YRRR', 3: 'YYRR', 4: 'YRYR', 5: 'YYYR', 6: 'YYYY'}


def purine_pyrimidine_classification(sequence_subsets, xarray=False):
    shorthand_purine_pyrimidine = shorthand_notation(purine_pyrimidine_sequence(sequence_subsets))
    shorthand_purine_pyrimidine_roll = np.array(roll_multiple(shorthand_purine_pyrimidine, 3)).T

    purine_pyrimidine_classification = np.zeros(len(shorthand_purine_pyrimidine_roll)).astype('int8')

    for index, pp_sequence in purine_pyrimidine_classes().items():
        purine_pyrimidine_classification[(shorthand_purine_pyrimidine_roll == pp_sequence).any(axis=1)] = index

    if xarray:
        purine_pyrimidine_classification = xr.DataArray(purine_pyrimidine_classification,
                                                        coords=dict(sequence_subset=sequence_subsets))

    return purine_pyrimidine_classification


def variant_score(variant):
    score = 0
    #     for i, l in enumerate(variant):
    #         if l == 'C':
    #             score += (i+1) * 1000
    #         elif l == 'T':
    #             score += (i+1) * 100
    #         elif l == 'G':
    #             score += (i+1) * 10
    #         elif l == 'A':
    #             score += (i+1) * 1
    order = {'C': 0, 'T': 1, 'A': 10, 'G': 11}
    for i, base in enumerate(variant):
        #         if i == 0 and base in ['T', 'C']:
        #             score -= 1
        if i == 0:
            score += order[base]
        #         elif i==3:
        #             score += order[::-1].index(base)
        elif i > 0:
            score += (order[previous_base] - order[base]) ** 3
        #         print(score)
        previous_base = base

    return score


def shorthand_sequence_rotationally_symmetric(sequence_subsets, xarray=False):
    invariant_shorthand = []
    for sequence_subset in sequence_subsets:
        variants = roll_multiple(shorthand_notation([sequence_subset]), 3)
        variant_scores = []
        for v in variants:
            #             print('test', variant_score(v[0]))
            variant_scores.append(variant_score(v[0]))
        selected_variant = variants[np.argmin(variant_scores)][0]
        invariant_shorthand.append(selected_variant)

    if xarray:
        invariant_shorthand = xr.DataArray(invariant_shorthand, coords=dict(sequence_subset=sequence_subsets))
    return invariant_shorthand

# def shorthand_sequence_rotationally_symmetric(sequence_subsets, xarray=False):
    # shorthand_purine_pyrimidine = shorthand_notation(purine_pyrimidine_sequence(sequence_subsets))
    # shorthand_purine_pyrimidine_roll = np.array(roll_multiple(shorthand_purine_pyrimidine, 3)).T
    # shorthand_purine_pyrimidine_roll

    # pps = np.array(list(purine_pyrimidine_classes().values()))

    # roll_indices = (shorthand_purine_pyrimidine_roll[:,:,None] == pps[None,None,:]).argmax(axis=-1).argmax(axis=-1)
    # shorthand_sequence_rotationally_symmetric = \
    #     np.array(roll_multiple(shorthand_notation(sequence_subsets), 3)).T[np.arange(len(roll_indices)),roll_indices]

    # if xarray:
    #     shorthand_sequence_rotationally_symmetric = xr.DataArray(shorthand_sequence_rotationally_symmetric,
    #                                                         coords=dict(sequence_subset=sequence_subsets))

    #     return shorthand_sequence_rotationally_symmetric


import sys
from PySide2.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QGridLayout, QTreeView, QApplication, QMainWindow, \
    QPushButton, QTabWidget, QTableWidget, QComboBox, QLineEdit
from PySide2.QtGui import QStandardItem, QStandardItemModel
from PySide2.QtCore import Qt

import numpy as np

class SelectionWidget(QWidget):
    def __init__(self, parent=None):
        super(SelectionWidget, self).__init__(parent)

        self.tree_view = QTreeView(self)
        self.model = QStandardItemModel()
        self.root = self.model.invisibleRootItem()
        self.model.setHorizontalHeaderLabels(['Variable', 'Channel', 'Aggregator', 'Operator', 'Threshold', 'Count'])
        self.tree_view.setModel(self.model)

        self.tree_view.setColumnWidth(0, 150)
        self.tree_view.setColumnWidth(1,100)

        self.model.itemChanged.connect(self.on_item_change)

        # variable_item = QStandardItem()
        # type_item = QStandardItem()
        # comparison_item = QStandardItem()
        # value_item = QStandardItem()
        # add_button_item = QStandardItem()
        # remove_button_item = QStandardItem()
        # self.root.appendRow([
        #     variable_item,
        #     type_item,
        #     comparison_item,
        #     value_item,
        #     add_button_item,
        #     remove_button_item,
        # ])
        # variable_item.setCheckable(True)

        # parentItem = self.root.child(self.root.rowCount() - 1)
        # testitem = QStandardItem('10')
        # selection1 = parentItem.appendRow([
        #     testitem,
        # ])



        #

        variable_combobox = QComboBox()
        variables = ['intensity', 'intensity_total', 'FRET']
        variable_combobox.addItems(variables)

        channel_combobox = QComboBox()
        channels = ['', '0', '1']
        channel_combobox.addItems(channels)

        aggregator_combobox = QComboBox()
        aggregators = ['mean', 'median', 'min', 'max']
        aggregator_combobox.addItems(aggregators)

        operator_combobox = QComboBox()
        operators = ['<', '>']
        operator_combobox.addItems(operators)

        threshold_lineedit = QLineEdit()

        add_button = QPushButton('Add')
        #
        # def add_function():
        #
        #     self.generate_selection(variable_combobox.currentText(),
        #                             channel_combobox.currentText(),
        #                             aggregator_combobox.currentText(),
        #                             operator_combobox.currentText(),
        #                             float(threshold_lineedit.text()))
        # add_button.clicked.connect(add_function)
        add_button.clicked.connect(self.add_selection)


        clear_button = QPushButton('Clear all')
        clear_button.clicked.connect(self.clear_selections)

        apply_to_selected_files_button = QPushButton('Apply to selected files')
        apply_to_selected_files_button.clicked.connect(self.apply_to_selected_files)


        self.add_selection_layout = QHBoxLayout()
        # self.add_selection_layout.addWidget(variable_combobox,1)
        # self.add_selection_layout.addWidget(channel_combobox,1)
        # self.add_selection_layout.addWidget(aggregator_combobox,1)
        # self.add_selection_layout.addWidget(operator_combobox,1)
        # self.add_selection_layout.addWidget(threshold_lineedit,1)
        self.add_selection_layout.addWidget(add_button)
        self.add_selection_layout.addWidget(clear_button)
        self.add_selection_layout.addWidget(apply_to_selected_files_button)

        selection_layout = QVBoxLayout()
        selection_layout.addWidget(self.tree_view)
        selection_layout.addLayout(self.add_selection_layout)

        self.setLayout(selection_layout)

        self.tree_view.setFixedWidth(700)
        #
        # self.add_button = QPushButton('Add')
        # self.add_button.clicked.connect(self.add_selection)
        # selection_layout = QVBoxLayout()
        # selection_layout.addWidget(self.tree_view)
        # selection_layout.addWidget(self.add_button)

        self.setLayout(selection_layout)

        self.update_final_selection = True
        self._file = None

    def on_item_change(self, item):
        if self.update_final_selection:
            selection_names = []
            for i in range(self.model.rowCount()):
                item = self.model.item(i)
                if item.checkState() == Qt.Checked:
                    selection_names.append(self.model.item(i).data())
            self.file.apply_selections(selection_names)
            self.refresh_selections()

    @property
    def file(self):
        return self._file

    @file.setter
    def file(self, file):
        self._file = file
        self.update_final_selection = False
        self.refresh_selections()
        self.update_final_selection = True
        # self.refresh_add_panel()

    def clear_selections(self):
        self.file.clear_selections()
        self.refresh_selections()

    def refresh_selections(self):
        self.root.removeRows(0, self.root.rowCount())
        if self.file is not None and '.nc' in self.file.extensions:
            self.setDisabled(False)
            for name, selection in self.file.selections_dataset.items():
                if not selection.attrs:
                    row_data = [name[10:], '', '', '', '']
                else:
                    columns = ['variable', 'channel', 'aggregator', 'operator', 'threshold']
                    row_data = [selection.attrs[c] for c in columns]
                row_data.append(selection.sum().item())
                items = [QStandardItem(str(d)) for d in row_data]
                items[0].setCheckable(True)
                items[0].setData(name)
                if 'selection_names' in self.file.selected.attrs.keys():
                    if np.isin(name, self.file.selected.attrs['selection_names']):
                        items[0].setCheckState(Qt.Checked)
                    else:
                        items[0].setCheckState(Qt.Unchecked)
                self.root.appendRow(items)

            items = [QStandardItem('') for _ in range(6)]
            self.root.appendRow(items)

            row_data = ['', '', '', '', 'Selected', str(self.file.number_of_selected_molecules)]
            items = [QStandardItem(str(d)) for d in row_data]
            self.root.appendRow(items)

            row_data = ['', '', '', '', 'Total', str(self.file.number_of_molecules)]
            items = [QStandardItem(str(d)) for d in row_data]
            self.root.appendRow(items)
        else:
            self.setDisabled(True)

    def add_selection(self):
        items = [QStandardItem(None) for _ in range(self.root.columnCount())]
        row_index = self.root.rowCount()-3
        # self.root.appendRow(items)
        self.root.insertRow(row_index, items)
        self.update_selection(row_index=row_index)

    def update_selection(self, row_index):
        i = row_index

        # row_items = self.root.takeRow(i)

        variable_item = self.root.child(i, 0)
        variable_combobox = QComboBox()
        variables = ['intensity', 'intensity_total', 'FRET']
        variable_combobox.addItems(variables)
        current_variable = variable_item.text()
        if current_variable != '':
            variable_combobox.setCurrentIndex(variables.index(variable_item.text()))
        self.tree_view.setIndexWidget(variable_item.index(), variable_combobox)

        channel_item = self.root.child(i, 1)
        channel_combobox = QComboBox()
        channels = ['', '0', '1']
        channel_combobox.addItems(channels)
        current_channel = channel_item.text()
        if current_channel != '':
            channel_combobox.setCurrentIndex(channels.index(channel_item.text()))
        self.tree_view.setIndexWidget(channel_item.index(), channel_combobox)

        aggregator_item = self.root.child(i, 2)
        aggregator_combobox = QComboBox()
        aggregators = ['mean', 'median', 'min', 'max']
        aggregator_combobox.addItems(aggregators)
        current_aggregator = aggregator_item.text()
        if current_aggregator != '':
            aggregator_combobox.setCurrentIndex(variables.index(aggregator_item.text()))
        self.tree_view.setIndexWidget(aggregator_item.index(), aggregator_combobox)

        operator_item = self.root.child(i, 3)
        operator_combobox = QComboBox()
        operators = ['<', '>']
        operator_combobox.addItems(operators)
        current_operator = operator_item.text()
        if current_operator != '':
            operator_combobox.setCurrentIndex(variables.index(operator_item.text()))
        self.tree_view.setIndexWidget(operator_item.index(), operator_combobox)

        threshold_item = self.root.child(i, 4)
        threshold_lineedit = QLineEdit()
        threshold_lineedit.setText(threshold_item.text())
        self.tree_view.setIndexWidget(threshold_item.index(), threshold_lineedit)

        apply_button_item = self.root.child(i, 5)
        apply_button = QPushButton('Apply')
        apply_function = lambda: self.generate_selection(variable_combobox.currentText(),
                                                         channel_combobox.currentText(),
                                                         aggregator_combobox.currentText(),
                                                         operator_combobox.currentText(),
                                                         float(threshold_lineedit.text()))
        apply_button.clicked.connect(apply_function)
        self.tree_view.setIndexWidget(apply_button_item.index(), apply_button)

        # remove_button_item = self.root.child(i, 5)
        # remove_button = QPushButton('Remove')
        # self.tree_view.setIndexWidget(remove_button_item.index(), remove_button)

    def apply_to_selected_files(self):
        self.file.copy_selections_to_selected_files()

    def generate_selection(self, variable, channel, aggregator, operator, threshold):

        # variable = variable.lower().replace(' ','_')
        # #TODO: Link these to available channels somehow
        # if variable[-6:] == '_green':
        #     channel = 0
        #     variable = variable[:-6]
        # elif variable[-4:] == '_red':
        #     channel = 1
        #     variable = variable[:-4]
        # else:
        #     channel = None

        self.file.add_selection(variable, channel, aggregator, operator, threshold)
        self.refresh_selections()



        # print(variable, ttype, operator, value)

            # variable_item = QStandardItem()
            # type_item = QStandardItem()
            # comparison_item = QStandardItem()
            # value_item = QStandardItem()
            # add_button_item = QStandardItem()
            # remove_button_item = QStandardItem()
            #
            #     variable_item,
            #     type_item,
            #     comparison_item,
            #     value_item,
            #     add_button_item,
            #     remove_button_item,
            # ])


        #     if selection.attrs
    #         name_item = QStandardItem(selection.selection.item()[10:].replace('_',' ').capitalize())
    #         count_item = QStandardItem(str(selection.sum('molecule').item()))
    #
    #         self.root.appendRow([
    #             name_item,
    #             count_item,
    #         ])
    # #
    # def refresh_add_panel(self):
    #     print('test')


        #
        # self.tree_view.expandAll()



## Old


# import sys
#
# # import contextlib
# # import io
# # import numpy as np
# # from pathlib import Path # For efficient path manipulation
# # import trace_analysis as ta
# # import matplotlib.pyplot as plt
# # import seaborn as sns
# # import tqdm
# # import time
# # import xarray as xr
# # from skimage.transform import AffineTransform
# # from tabulate import tabulate
#
# # from trace_analysis.plugins.holliday_junction.holliday_junction import *
# import sys
# from PySide2.QtWidgets import (QApplication, QMainWindow, QTreeView, QListView, QFileSystemModel,
#                                QVBoxLayout, QHBoxLayout, QWidget, QStyledItemDelegate, QComboBox)
# from PySide2.QtCore import Qt
# from PySide2.QtGui import QStandardItem, QStandardItemModel
# from PySide2.QtWidgets import QTreeWidget, QTreeWidgetItem, QComboBox, QTextEdit, QLineEdit, QPushButton
#
# class TreeViewExample(QMainWindow):
#     def __init__(self):
#         super(TreeViewExample, self).__init__()
#
#         self.setWindowTitle("QTreeView Example")
#         self.setGeometry(100, 100, 800, 600)
#
#         # Create a QTreeView and set up a QFileSystemModel as the model
#
#         # # Set the custom delegate for a specific column (e.g., column 1)
#         # child1 = QTreeWidgetItem(["Field 1", "<Free Text>"])
#         # treeWidget = QTreeWidget(self)
#         # treeWidget.setItemWidget(child1, 1, QComboBox())
#         # self.root.addChild(child1)
#
#         self.selection = SelectionWidget()
#
#         # Set up the layout
#         layout = QVBoxLayout(self)
#         layout.addWidget(self.selection)
#
#         central_widget = QWidget(self)
#         central_widget.setLayout(layout)
#         self.setCentralWidget(central_widget)
#
# class SelectionWidget(QWidget):
#     def __init__(self, parent=None):
#         super(SelectionWidget, self).__init__(parent)
#
#         self.tree_view = QTreeView(self)
#         self.model = QStandardItemModel()
#         self.root = self.model.invisibleRootItem()
#         self.model.setHorizontalHeaderLabels(['Variable', 'Channel', 'Aggregator', 'Operator', 'Threshold', 'Count'])
#         self.tree_view.setModel(self.model)
#
#         self.tree_view.setColumnWidth(0, 150)
#         self.tree_view.setColumnWidth(1,100)
#
#
#         # variable_item = QStandardItem()
#         # type_item = QStandardItem()
#         # comparison_item = QStandardItem()
#         # value_item = QStandardItem()
#         # add_button_item = QStandardItem()
#         # remove_button_item = QStandardItem()
#         # self.root.appendRow([
#         #     variable_item,
#         #     type_item,
#         #     comparison_item,
#         #     value_item,
#         #     add_button_item,
#         #     remove_button_item,
#         # ])
#         # variable_item.setCheckable(True)
#
#         # parentItem = self.root.child(self.root.rowCount() - 1)
#         # testitem = QStandardItem('10')
#         # selection1 = parentItem.appendRow([
#         #     testitem,
#         # ])
#
#
#
#         #
#
#         variable_combobox = QComboBox()
#         variables = ['intensity', 'intensity_total', 'FRET']
#         variable_combobox.addItems(variables)
#
#         channel_combobox = QComboBox()
#         channels = ['', '0', '1']
#         channel_combobox.addItems(channels)
#
#         aggregator_combobox = QComboBox()
#         aggregators = ['mean', 'median', 'min', 'max']
#         aggregator_combobox.addItems(aggregators)
#
#         operator_combobox = QComboBox()
#         operators = ['<', '>']
#         operator_combobox.addItems(operators)
#
#         threshold_lineedit = QLineEdit()
#
#         add_button = QPushButton('Add')
#
#         def add_function():
#
#             self.generate_selection(variable_combobox.currentText(),
#                                     channel_combobox.currentText(),
#                                     aggregator_combobox.currentText(),
#                                     operator_combobox.currentText(),
#                                     float(threshold_lineedit.text()))
#         add_button.clicked.connect(add_function)
#
#         clear_button = QPushButton('Clear all')
#         clear_button.clicked.connect(self.clear_selections)
#
#
#         self.add_selection_layout = QHBoxLayout()
#         self.add_selection_layout.addWidget(variable_combobox,1)
#         self.add_selection_layout.addWidget(channel_combobox,1)
#         self.add_selection_layout.addWidget(aggregator_combobox,1)
#         self.add_selection_layout.addWidget(operator_combobox,1)
#         self.add_selection_layout.addWidget(threshold_lineedit,1)
#         self.add_selection_layout.addWidget(add_button,0.5)
#         self.add_selection_layout.addWidget(clear_button,0.5)
#
#         selection_layout = QVBoxLayout()
#         selection_layout.addWidget(self.tree_view)
#         selection_layout.addLayout(self.add_selection_layout)
#
#         self.setLayout(selection_layout)
#
#         self.tree_view.setFixedWidth(700)
#         #
#         # self.add_button = QPushButton('Add')
#         # self.add_button.clicked.connect(self.add_selection)
#         # selection_layout = QVBoxLayout()
#         # selection_layout.addWidget(self.tree_view)
#         # selection_layout.addWidget(self.add_button)
#
#         self.setLayout(selection_layout)
#
#         self._file = None
#
#     @property
#     def file(self):
#         return self._file
#
#     @file.setter
#     def file(self, file):
#         self._file = file
#         self.refresh_selections()
#         # self.refresh_add_panel()
#
#     def clear_selections(self):
#         self.file.clear_selections()
#         self.refresh_selections()
#
#     def refresh_selections(self):
#         self.root.removeRows(0,self.root.rowCount())
#         for name, selection in self.file.selections_dataset.items():
#             if not selection.attrs:
#                 row_data = [name[10:], '', '', '', '']
#             else:
#                 columns = ['variable', 'channel', 'aggregator', 'operator', 'threshold']
#                 row_data = [selection.attrs[c] for c in columns]
#             row_data.append(selection.sum().item())
#             items = [QStandardItem(str(d)) for d in row_data]
#             self.root.appendRow(items)
#             items[0].setCheckable(True)
#
#     def add_selection(self):
#         items = [QStandardItem(None) for _ in range(self.root.columnCount())]
#         self.root.appendRow(items)
#         self.update_selection(row_index=self.root.rowCount()-1)
#
#     def update_selection(self, row_index):
#         i = row_index
#
#         # row_items = self.root.takeRow(i)
#
#         variable_item = self.root.child(i, 0)
#         variable_combobox = QComboBox()
#         variables = ['intensity', 'intensity_total', 'FRET']
#         variable_combobox.addItems(variables)
#         current_variable = variable_item.text()
#         if current_variable != '':
#             variable_combobox.setCurrentIndex(variables.index(variable_item.text()))
#         self.tree_view.setIndexWidget(variable_item.index(), variable_combobox)
#
#         channel_item = self.root.child(i, 1)
#         channel_combobox = QComboBox()
#         channels = ['', '0', '1']
#         channel_combobox.addItems(variables)
#         current_channel = channel_item.text()
#         if current_channel != '':
#             channel_combobox.setCurrentIndex(channels.index(channel_item.text()))
#         self.tree_view.setIndexWidget(channel_item.index(), channel_combobox)
#
#         aggregator_item = self.root.child(i, 2)
#         aggregator_combobox = QComboBox()
#         aggregators = ['mean', 'median', 'min', 'max']
#         aggregator_combobox.addItems(aggregators)
#         current_aggregator = aggregator_item.text()
#         if current_aggregator != '':
#             aggregator_combobox.setCurrentIndex(variables.index(aggregator_item.text()))
#         self.tree_view.setIndexWidget(aggregator_item.index(), aggregator_combobox)
#
#         operator_item = self.root.child(i, 3)
#         operator_combobox = QComboBox()
#         operators = ['<', '>']
#         operator_combobox.addItems(operators)
#         current_operator = operator_item.text()
#         if current_operator != '':
#             operator_combobox.setCurrentIndex(variables.index(operator_item.text()))
#         self.tree_view.setIndexWidget(operator_item.index(), operator_combobox)
#
#         threshold_item = self.root.child(i, 4)
#         threshold_lineedit = QLineEdit()
#         threshold_lineedit.setText(threshold_item.text())
#         self.tree_view.setIndexWidget(threshold_item.index(), threshold_lineedit)
#
#         apply_button_item = self.root.child(i, 5)
#         apply_button = QPushButton('Apply')
#         apply_function = lambda: self.generate_selection(variable_combobox.currentText(),
#                                                          aggregator_combobox.currentText(),
#                                                          channel_combobox.currentText(),
#                                                          operator_combobox.currentText(),
#                                                          float(threshold_lineedit.text()))
#         apply_button.clicked.connect(apply_function)
#         self.tree_view.setIndexWidget(apply_button_item.index(), apply_button)
#
#         # remove_button_item = self.root.child(i, 5)
#         # remove_button = QPushButton('Remove')
#         # self.tree_view.setIndexWidget(remove_button_item.index(), remove_button)
#
#     def generate_selection(self, variable, channel, aggregator, operator, threshold):
#
#         # variable = variable.lower().replace(' ','_')
#         # #TODO: Link these to available channels somehow
#         # if variable[-6:] == '_green':
#         #     channel = 0
#         #     variable = variable[:-6]
#         # elif variable[-4:] == '_red':
#         #     channel = 1
#         #     variable = variable[:-4]
#         # else:
#         #     channel = None
#
#         self.file.add_selection(variable, channel, aggregator, operator, threshold)
#         self.refresh_selections()
#
#
#
#         # print(variable, ttype, operator, value)
#
#             # variable_item = QStandardItem()
#             # type_item = QStandardItem()
#             # comparison_item = QStandardItem()
#             # value_item = QStandardItem()
#             # add_button_item = QStandardItem()
#             # remove_button_item = QStandardItem()
#             #
#             #     variable_item,
#             #     type_item,
#             #     comparison_item,
#             #     value_item,
#             #     add_button_item,
#             #     remove_button_item,
#             # ])
#
#
#         #     if selection.attrs
#     #         name_item = QStandardItem(selection.selection.item()[10:].replace('_',' ').capitalize())
#     #         count_item = QStandardItem(str(selection.sum('molecule').item()))
#     #
#     #         self.root.appendRow([
#     #             name_item,
#     #             count_item,
#     #         ])
#     # #
#     # def refresh_add_panel(self):
#     #     print('test')
#
#
#         #
#         # self.tree_view.expandAll()
#
#
#
# if __name__ == "__main__":
#     app = QApplication(sys.argv)
#     window = TreeViewExample()
#     window.show()
#
#     from trace_analysis import Experiment
#     test_path = r'C:\Users\ivoseverins\surfdrive\Promotie\Code\Python\traceAnalysis\twoColourExampleData\BN_TIRF'
#     exp = Experiment(test_path)
#
#     window.selection.file = exp.files[-1]
#
#     sys.exit(app.exec_())



import papylio as pp

# Experiment import
exp = pp.Experiment()
exp.print_files()


# Emission channel mapping
mapping_file_index = 0
mapping_file = exp.files[mapping_file_index]

mapping_file.perform_mapping()

figure, axis = mapping_file.show_image()
mapping_file.mapping.show(axis=axis, show_source=True)


# File selection
file_selection = exp.files[2:10]  # Selection of files: change to the desired file range


# Background subtraction
file_selection.movie.determine_spatial_background_correction()


# Molecule localization - test on one file
test_file = file_selection[0]  # Change number to the desired file number
test_file.find_coordinates()
test_file.show_coordinates_in_image()


# Molecule localization - apply to all files
file_selection.find_coordinates()


# Trace extraction
file_selection.extract_traces()


# Show traces
test_file = file_selection[0]  # Change number to the desired file number
test_file.show_traces(plot_variables=['intensity', 'FRET'],
                     ylims=[(0, 35000), (0, 1)],
                     colours=[('green', 'red'), ('blue')])
